#!/usr/bin/env python
# encoding: utf-8

from .parts.honeycombpart import HoneycombPart
from .parts.squarepart import SquarePart
from .parts.part import Part
from .strand import Strand
from operator import itemgetter
import cadnano2.util as util
import cadnano2.cadnano as cadnano
util.qtWrapImport('QtCore', globals(), ['pyqtSignal', 'QObject'])
util.qtWrapImport('QtGui', globals(), ['QUndoCommand', 'QUndoStack'])


class Document(QObject):
    """
    The Document class is the root of the model. It has two main purposes:
    1. Serve as the parent all Part objects within the model.
    2. Track all sub-model actions on its undoStack.
    """
    def __init__(self):
        super(Document, self).__init__()
        self._undoStack = QUndoStack()
        self._parts = []
        self._assemblies = []
        self._controller = None
        self._selectedPart = None
        # the dictionary maintains what is selected
        self._selectionDict = {}
        # the added list is what was recently selected or deselected
        self._selectedChangedDict = {}
        cadnano.app().documentWasCreatedSignal.emit(self)

    ### SIGNALS ###
    documentPartAddedSignal = pyqtSignal(object, QObject)  # doc, part

    # dict of tuples of objects using the reference as the key,
    # and the value is a tuple with meta data
    # in the case of strands the metadata would be which endpoints of selected
    # e.g. { objectRef: (value0, value1),  ...}
    documentSelectedChangedSignal = pyqtSignal(dict)  # tuples of items + data
    documentSelectionFilterChangedSignal = pyqtSignal(list) # doc, filterlist

    documentViewResetSignal = pyqtSignal(QObject)
    documentClearSelectionsSignal = pyqtSignal(QObject)


    ### SLOTS ###

    ### ACCESSORS ###
    def undoStack(self):
        """
        This is the actual undoStack to use for all commands. Any children
        needing to perform commands should just ask their parent for the
        undoStack, and eventually the request will get here.
        """
        return self._undoStack

    def parts(self):
        """Returns a list of parts associated with the document."""
        return self._parts

    def assemblies(self):
        """Returns a list of assemblies associated with the document."""
        return self._assemblies

    ### PUBLIC METHODS FOR QUERYING THE MODEL ###
    def selectedPart(self):
        return self._selectedPart

    def addToSelection(self, obj, value):
        self._selectionDict[obj] = value
        self._selectedChangedDict[obj] = value
    # end def

    def removeFromSelection(self, obj):
        if obj in self._selectionDict:
            del self._selectionDict[obj]
            self._selectedChangedDict[obj] = (False, False)
            return True
        else:
            return False
    # end def

    def clearSelections(self):
        """
        Only clear the dictionary
        """
        self._selectionDict = {}
    # end def

    def addStrandToSelection(self, strand, value):
        sS = strand.strandSet()
        if sS in self._selectionDict:
            self._selectionDict[sS][strand] = value
        else:
            self._selectionDict[sS] = {strand: value}
        self._selectedChangedDict[strand] = value
    # end def

    def removeStrandFromSelection(self, strand):
        sS = strand.strandSet()
        if sS in self._selectionDict:
            temp = self._selectionDict[sS]
            if strand in temp:
                del temp[strand]
                if len(temp) == 0:
                    del self._selectionDict[sS]
                self._selectedChangedDict[strand] = (False, False)
                return True
            else:
                return False
        else:
            return False
    # end def

    def selectionDict(self):
        return self._selectionDict
    # end def
    
    def selectedOligos(self):
        """
        as long as one endpoint of a strand is in the selection, then the oligo
        is considered selected
        """
        sDict = self._selectionDict
        selectedOs = set()
        for sS in sDict.keys():
            for strand in sS:
                 selectedOs.add(strand.oligo())
            # end for
        # end for
        return selectedOs if len(selectedOs) > 0 else None
    #end def
    
    def clearAllSelected(self):
        self._selectionDict = {}
        # the added list is what was recently selected or deselected
        self._selectedChangedDict = {}
        self.documentClearSelectionsSignal.emit(self)
    # end def

    def isModelSelected(self, obj):
        return obj in self._selectionDict
    # end def

    def isModelStrandSelected(self, strand):
        sS = strand.strandSet()
        if sS in self._selectionDict:
            if strand in self._selectionDict[sS]:
                return True
            else:
                return False
        else:
            return False
    # end def

    def getSelectedValue(self, obj):
        """
        obj is an objects to look up
        it is prevetted to be in the dictionary
        """
        return self._selectionDict[obj]

    def getSelectedStrandValue(self, strand):
        """
        strand is an objects to look up
        it is prevetted to be in the dictionary
        """
        return self._selectionDict[strand.strandSet()][strand]
    # end def

    def sortedSelectedStrands(self, strandSet):
        # outList = self._selectionDict[strandSet].keys()
        # outList.sort(key=Strand.lowIdx)
        outList = list(self._selectionDict[strandSet].items())
        getLowIdx = lambda x: Strand.lowIdx(itemgetter(0)(x))
        outList.sort(key=getLowIdx)
        return outList
    # end def

    def determineStrandSetBounds(self, selectedStrandList, strandSet):
        minLowDelta = strandSet.partMaxBaseIdx()
        minHighDelta = strandSet.partMaxBaseIdx()  # init the return values
        sSDict = self._selectionDict[strandSet]
        # get the StrandSet index of the first item in the list
        sSIdx = strandSet._findIndexOfRangeFor(selectedStrandList[0][0])[2]
        sSList = strandSet._strandList
        lenSSList = len(sSList)
        maxSSIdx = lenSSList - 1
        i = 0
        for strand, value in selectedStrandList:
            while strand != sSList[sSIdx]:
                # incase there are gaps due to double xovers
                sSIdx += 1
            # end while
            idxL, idxH = strand.idxs()
            if value[0]:    # the end is selected
                if sSIdx > 0:
                    lowNeighbor = sSList[sSIdx - 1]
                    if lowNeighbor in sSDict:
                        valueN = sSDict[lowNeighbor]
                        # we only care if the low neighbor is not selected
                        temp = minLowDelta if valueN[1] \
                                        else idxL - lowNeighbor.highIdx() - 1
                    # end if
                    else:  # not selected
                        temp = idxL - lowNeighbor.highIdx() - 1
                    # end else
                else:
                    temp = idxL - 0
                # end else
                if temp < minLowDelta:
                    minLowDelta = temp
                # end if
                # check the other end of the strand
                if not value[1]:
                    temp = idxH - idxL - 1
                    if temp < minHighDelta:
                        minHighDelta = temp
            # end if
            if value[1]:
                if sSIdx < maxSSIdx:
                    highNeighbor = sSList[sSIdx + 1]
                    if highNeighbor in sSDict:
                        valueN = sSDict[highNeighbor]
                        # we only care if the low neighbor is not selected
                        temp = minHighDelta if valueN[0] \
                                        else highNeighbor.lowIdx() - idxH - 1
                    # end if
                    else:  # not selected
                        temp = highNeighbor.lowIdx() - idxH - 1
                    # end else
                else:
                    temp = strandSet.partMaxBaseIdx() - idxH
                # end else
                if temp < minHighDelta:
                    minHighDelta = temp
                # end if
                # check the other end of the strand
                if not value[0]:
                    temp = idxH - idxL - 1
                    if temp < minLowDelta:
                        minLowDelta = temp
            # end if
            # increment counter
            sSIdx += 1
        # end for
        return (minLowDelta, minHighDelta)
    # end def

    def getSelectionBounds(self):
        minLowDelta = -1
        minHighDelta = -1
        for strandSet in self._selectionDict.keys():
            selectedList = self.sortedSelectedStrands(strandSet)
            tempLow, tempHigh = self.determineStrandSetBounds(
                                                    selectedList, strandSet)
            if tempLow < minLowDelta or minLowDelta < 0:
                minLowDelta = tempLow
            if tempHigh < minHighDelta or minHighDelta < 0:
                minHighDelta = tempHigh
        # end for Mark train bus to metro
        return (minLowDelta, minHighDelta)
    # end def

    # def operateOnStrandSelection(self, method, arg, both=False):
    #     pass
    # # end def

    def deleteSelection(self, useUndoStack=True):
        """
        Delete selected strands. First iterates through all selected strands
        and extracts refs to xovers and strands. Next, calls removeXover
        on xoverlist as part of its own macroed command for isoluation
        purposes. Finally, calls removeStrand on all strands that were 
        fully selected (low and high), or had at least one non-xover
        endpoint selected.
        """
        xoList = []
        strandDict = {}
        for strandSetDict in list(self._selectionDict.values()):
            for strand, selected in list(strandSetDict.items()):
                part = strand.virtualHelix().part()
                idxL, idxH = strand.idxs()
                strand5p = strand.connection5p()
                strand3p = strand.connection3p()
                # both ends are selected
                strandDict[strand] = selected[0] and selected[1]

                # only look at 3' ends to handle xover deletion
                sel3p = selected[0] if idxL == strand.idx3Prime() else selected[1]
                if sel3p:  # is idx3p selected?
                    if strand3p:  # is there an xover
                        xoList.append((part, strand, strand3p, useUndoStack))
                    else:  # idx3p is a selected endpoint
                        strandDict[strand] = True
                else:
                    if not strand5p:  # idx5p is a selected endpoint
                        strandDict[strand] = True


        if useUndoStack and xoList:
            self.undoStack().beginMacro("Delete xovers")
        for part, strand, strand3p, useUndo in xoList:
            Part.removeXover(part, strand, strand3p, useUndo)
            self.removeStrandFromSelection(strand)
            self.removeStrandFromSelection(strand3p)
        self._selectionDict = {}
        self.documentClearSelectionsSignal.emit(self)
        if useUndoStack:
            if xoList: # end xover macro if it was started
                self.undoStack().endMacro()
            if True in list(strandDict.values()):
                self.undoStack().beginMacro("Delete selection")
            else:
                return  # nothing left to do
        for strand, delete in list(strandDict.items()):
            if delete:
                strand.strandSet().removeStrand(strand)
        if useUndoStack:
            self.undoStack().endMacro()

    def paintSelection(self, scafColor, stapColor, useUndoStack=True):
        """Delete xovers if present. Otherwise delete everything."""
        scafOligos = {}
        stapOligos = {}
        for strandSetDict in list(self._selectionDict.values()):
            for strand, value in list(strandSetDict.items()):
                if strand.isScaffold():
                    scafOligos[strand.oligo()] = True
                else:
                    stapOligos[strand.oligo()] = True

        if useUndoStack:
            self.undoStack().beginMacro("Paint strands")
        for olg in list(scafOligos.keys()):
            olg.applyColor(scafColor)
        for olg in list(stapOligos.keys()):
            olg.applyColor(stapColor)
        if useUndoStack:
            self.undoStack().endMacro()

    def resizeSelection(self, delta, useUndoStack=True):
        """
        Moves the selected idxs by delta by first iterating over all strands
        to calculate new idxs (method will return if snap-to behavior would
        create illegal state), then applying a resize command to each strand.
        """
        resizeList = []

        # calculate new idxs
        for strandSetDict in self._selectionDict.values():
            for strand, selected in strandSetDict.items():
                part = strand.virtualHelix().part()
                idxL, idxH = strand.idxs()
                newL, newH = strand.idxs()
                deltaL = deltaH = delta

                # process xovers to get revised delta
                if selected[0] and strand.connectionLow():
                    newL = part.xoverSnapTo(strand, idxL, delta)
                    if newL == None:
                        return
                    deltaH = newL-idxL
                if selected[1] and strand.connectionHigh():
                    newH = part.xoverSnapTo(strand, idxH, delta)
                    if newH == None:
                        return
                    deltaL = newH-idxH

                # process endpoints
                if selected[0] and not strand.connectionLow():
                    newL = idxL + deltaL
                if selected[1] and not strand.connectionHigh():
                    newH = idxH + deltaH

                if newL > newH:  # check for illegal state
                    return

                resizeList.append((strand, newL, newH))
            # end for
        # end for

        # execute the resize commands
        if useUndoStack:
            self.undoStack().beginMacro("Resize Selection")

        for strand, idxL, idxH in resizeList:
            Strand.resize(strand, (idxL, idxH), useUndoStack)

        if useUndoStack:
            self.undoStack().endMacro()
    # end def

    def updateSelection(self):
        """
        do it this way in the future when we have
        a better signaling architecture between views
        """
        # self.documentSelectedChangedSignal.emit(self._selectedChangedDict)
        """
        For now, individual objects need to emit signals
        """
        for obj, value in self._selectedChangedDict.items():
            obj.selectedChangedSignal.emit(obj, value)
        # end for
        self._selectedChangedDict = {}
        # for sS in self._selectionDict:
        #     print self.sortedSelectedStrands(sS)
    # end def

    def resetViews(self):
        # This is a fast way to clear selections and the views.
        # We could manually deselect each item from the Dict, but we'll just
        # let them be garbage collect
        # the dictionary maintains what is selected
        self._selectionDict = {}
        # the added list is what was recently selected or deselected
        self._selectedChangedDict = {}
        self.documentViewResetSignal.emit(self)
    # end def

    ### PUBLIC METHODS FOR EDITING THE MODEL ###
    def addHoneycombPart(self):
        """
        Create and store a new DNAPart and instance, and return the instance.
        """
        dnapart = None
        if len(self._parts) == 0:
            dnapart = HoneycombPart(document=self)
            self._addPart(dnapart)
        return dnapart

    def addSquarePart(self):
        """
        Create and store a new DNAPart and instance, and return the instance.
        """
        dnapart = None
        if len(self._parts) == 0:
            dnapart = SquarePart(document=self)
            self._addPart(dnapart)
        return dnapart

    def removeAllParts(self):
        """Used to reset the document. Not undoable."""
        self.documentClearSelectionsSignal.emit(self)
        for part in self._parts:
            part.remove(useUndoStack=False)
    # end def

    def removePart(self, part):
        self.documentClearSelectionsSignal.emit(self)
        self._parts.remove(part)
        

    ### PUBLIC SUPPORT METHODS ###
    def controller(self):
        return self._controller

    def setController(self, controller):
        """Called by DocumentController setDocument method."""
        self._controller = controller
    # end def

    def setSelectedPart(self, newPart):
        if self._selectedPart == newPart:
            return
        self._selectedPart = newPart
    # end def

    ### PRIVATE SUPPORT METHODS ###
    def _addPart(self, part, useUndoStack=True):
        """Add part to the document via AddPartCommand."""
        c = self.AddPartCommand(self, part)
        util.execCommandList(
                        self, [c], desc="Add part", useUndoStack=useUndoStack)
        return c.part()
    # end def

    ### COMMANDS ###
    class AddPartCommand(QUndoCommand):
        """
        Undo ready command for deleting a part.
        """
        def __init__(self, document, part):
            QUndoCommand.__init__(self)
            self._doc = document
            self._part = part
        # end def

        def part(self):
            return self._part
        # end def

        def redo(self):
            if len(self._doc._parts) == 0:
                self._doc._parts.append(self._part)
                self._part.setDocument(self._doc)
                self._doc.setSelectedPart(self._part)
                self._doc.documentPartAddedSignal.emit(self._doc, self._part)
        # end def

        def undo(self):
            self._doc.removePart(self._part)
            self._part.setDocument(None)
            self._doc.setSelectedPart(None)
            self._part.partRemovedSignal.emit(self._part)
            # self._doc.documentPartAddedSignal.emit(self._doc, self._part)
        # end def
    # end class
# end class



# Copyright 2011 Autodesk, Inc.  All rights reserved.
#
# The MIT License
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
# http://www.opensource.org/licenses/mit-license.php

# Example Code:
# from maya import cmds
# cmds.loadPlugin("removedMsgCmd.py")
# cmds.spRemovedMsg()

"""
removedMsgCmd.py
Created by Simon Breslav on 2011/09/06
A command that adds a callback function when Maya Nodes are deleted
"""

import sys
import maya.OpenMaya as OpenMaya
import maya.OpenMayaMPx as OpenMayaMPx
from controllers.mayacontrollers.mayaObjectManager import Mom

kPluginCmdName = "spRemovedMsg"
messageId = 0
messageIdSet = False


def removeCallback(id):
    """
    Function to remove the callback function from Maya, given the ID of the
    function, the id is created in createParentRemovedCallback function
    """
    try:
        OpenMaya.MMessage.removeCallback(id)
    except:
        sys.stderr.write("Failed to remove callback\n")


def dagParentRemovedCallback(child, parent, clientData):
    """
    Callback function that removes the strands from the model if the
    3D strand is delete in 3D view.
    """
    mom = Mom()
    children = child.fullPathName().split("|")
    for c in children:
        if c.startswith(mom.helixMeshName):
            if c in mom.mayaToCn:
                strand = mom.mayaToCn[c]
                if strand:
                    # print "Strand %s : %s needs removal" % (c, strand)
                    mID = mom.strandMayaID(strand)
                    mom.removeIDMapping(mID, strand)
                    strand.strandSet().removeStrand(strand)
                else:
                    print("Error: no Strand inside mayaObjectModel")
            else:
                pass
                # print "dagParentRemovedCallback: %s already deleted" % c
        elif c.startswith(mom.decoratorMeshName):
            if c in mom.decoratorToVirtualHelixItem:
                pass
                #decoratorObject = mom.decoratorToVirtualHelixItem[ c ]
                #virtualHelixItem = decoratorObject[0]
                #virtualHelixItem.updateDecorators()


def createParentRemovedCallback(stringData):
    """
    Function that creates the actual callback function and returns its ID
    """
    # global declares module level variables that will be assigned
    global messageIdSet
    try:
        id = OpenMaya.MDagMessage.addParentRemovedCallback( \
                                    dagParentRemovedCallback, stringData)
    except:
        sys.stderr.write("Failed to install dag parent removed callback\n")
        messageIdSet = False
    else:
        messageIdSet = True
    return id


# command
class scriptedCommand(OpenMayaMPx.MPxCommand):
    """
    Maya Command that adds a callback function when something is removed
    """
    def __init__(self):
        """Initialize the commands"""
        OpenMayaMPx.MPxCommand.__init__(self)

    def doIt(self, argList):
        """Execute the command"""
        global messageId
        if (messageIdSet):
            print("Message callback already installed")
        else:
            print("Installing parent removed callback message")
            messageId = createParentRemovedCallback("_noData_")


# Creator
def cmdCreator():
    """Wrapper function that created the command"""
    return OpenMayaMPx.asMPxPtr(scriptedCommand())

# Initialize the script plug-in
def initializePlugin(mobject):
    """Register the command with Maya""" 
    mplugin = OpenMayaMPx.MFnPlugin(mobject)
    try:
        mplugin.registerCommand(kPluginCmdName, cmdCreator)
    except:
        sys.stderr.write("Failed to register command: %s\n" % name)
        raise


# Uninitialize the script plug-in
def uninitializePlugin(mobject):
    """Remove the command and the callback""" 
    # Remove the callback
    if (messageIdSet):
        removeCallback(messageId)
    # Remove the plug-in command
    mplugin = OpenMayaMPx.MFnPlugin(mobject)
    try:
        mplugin.deregisterCommand(kPluginCmdName)
    except:
        sys.stderr.write("Failed to unregister command: %s\n" % kPluginCmdName)
        raise



import os
from cadnano2.cadnano import app
from .selecttool import SelectTool
from .penciltool import PencilTool
from .breaktool import BreakTool
from .erasetool import EraseTool
from .insertiontool import InsertionTool
from .skiptool import SkipTool
from .painttool import PaintTool
from .addseqtool import AddSeqTool
import cadnano2.util as util

# import Qt stuff into the module namespace with PySide, PyQt4 independence
util.qtWrapImport('QtCore', globals(), ['QObject', 'pyqtSignal'])
util.qtWrapImport('QtGui', globals(), [ 'QActionGroup'])


class PathToolManager(QObject):
    """
    Manages the interactions between Path widgets / UI elements and the model.
    """
    def __init__(self, win):
        super(PathToolManager, self).__init__()
        self.window = win
        self._activeTool = None
        self._activePart = None
        self.selectTool = SelectTool(self)
        self.pencilTool = PencilTool(self)
        self.breakTool = BreakTool(self)
        self.eraseTool = EraseTool(self)
        self.insertionTool = InsertionTool(self)
        self.skipTool = SkipTool(self)
        self.paintTool = PaintTool(self) # (self, win.pathGraphicsView.toolbar)
        self.addSeqTool = AddSeqTool(self)

        def installTool(toolName, window):
            toolWidget = getattr(window, 'actionPath' + toolName)
            lToolName = toolName[0].lower() + toolName[1:]
            tool = getattr(self, lToolName + 'Tool')
            tool.actionName = 'actionPath' + toolName

            def clickHandler(self):
                toolWidget.setChecked(True)
                self.setActiveTool(tool)
                if hasattr(tool, 'widgetClicked'):
                    tool.widgetClicked()
            selectToolMethodName = 'choose' + toolName + 'Tool'
            setattr(self.__class__, selectToolMethodName, clickHandler)
            handler = getattr(self, selectToolMethodName)
            toolWidget.triggered.connect(handler)
            return toolWidget

        tools = ('Select', 'Pencil', 'Break', 'Erase', 'Insertion', 'Skip', 'Paint', 'AddSeq')
        ag = QActionGroup(win)
        # Call installTool on every tool
        list(map((lambda toolName: ag.addAction(installTool(toolName, win))), tools))
        ag.setExclusive(True)
        # Select the preferred Startup tool
        startupToolName = app().prefs.getStartupToolName()
        getattr(self, 'choose' + startupToolName + 'Tool')()

    ### SIGNALS ###
    activeToolChangedSignal = pyqtSignal(str)

    ### SLOTS ###

    ### METHODS ###
    def activePart(self):
        return self._activePart

    def setActivePart(self, part):
        self._activePart = part

    def activeTool(self):
        return self._activeTool

    def setActiveTool(self, newActiveTool):
        if newActiveTool == self._activeTool:
            return
        if self.lastLocation():
            newActiveTool.updateLocation(*self.lastLocation())
        if self._activeTool:
            self._activeTool.setActive(False)
        if str(newActiveTool) == "selectTool":
            self.window.activateSelection(True)
        elif str(newActiveTool) == "paintTool":
            self.window.activateSelection(True)
        else:
            self.window.activateSelection(False)
        self._activeTool = newActiveTool
        self._activeTool.setActive(True)
        self.activeToolChangedSignal.emit(self._activeTool.actionName)

    def isSelectToolActive(self):
        if self.activeTool() == self.selectTool:
            return True
        return False

    def lastLocation(self):
        """(PathHelix, posInScene) or None, depending on where
        the mouse is (basically, pathHelix and position of
        the last event seen by the active tool)"""
        if self._activeTool == None:
            return None
        return self._activeTool.lastLocation()
# end class



#!/usr/bin/env python
# encoding: utf-8

from .fetchfile import fetchFile
import shutil
from os.path import basename, dirname, splitext, exists, split, abspath
import os

# list of tuples of the url and the subdirectory to copy to the include directory
urls = [('http://pypi.python.org/packages/source/n/networkx/networkx-1.6.tar.gz', 'networkx')
        ]

# this is intended to be run in the include folder of cadnano2
if __name__ == '__main__':
    for urlBlock in urls:
        url, subfolder_name = urlBlock
        filename = basename(url)
        base_url = dirname(url)
        folder = fetchFile(filename, base_url)
        this_dirname, this_filename = split(abspath(__file__))
        os.chdir(this_dirname)
        if subfolder_name:
            subfolder_path = folder + '/' + subfolder_name
            # remove existing folder
            if os.path.exists(subfolder_name):
                print("file exists")
                shutil.rmtree(subfolder_name)
            else:
                print("file does not exist", subfolder_name)
            shutil.move(subfolder_path, subfolder_name)
            shutil.rmtree(folder)
    # end for



class QUndoCommand(object):
    children = []
    name = "untitled"
    def undo(self):
        for c in reversed(self.children):
            c.undo()
    def redo(self):
        for c in self.children:
            c.redo()

class QUndoStack(object):
    undoCmds = []
    macroStack = []  # list of lists
    macroNameStack = []
    index = 0
    clean = True
    def isClean(self):
        return self.clean
    def setClean(self):
        self.clean = True
    def clean(self):
        self.clean = True
        self.undoCmds = []
        self.index = 0
    def beginMacro(self, macroName):
        self.macroStack.append([])
        self.macroNameStack.append(macroName)
        self.clean = False
    def push(self, cmd):
        if self.macroStack:
            self.macroStack[-1].append(cmd)
        else:
            self.undoCmds.append(cmd)
            self.index = len(self.undoCmds) - 1
        cmd.redo()
        self.clean = False
    def endMacro(self):
        l = len(self.macroStack)
        if l == 0:
            assert(False)  # Can't end a macro that wasn't begun
        cmd = QUndoCommand()
        cmd.children = self.macroStack.pop()
        cmd.name = self.macroNameStack.pop()
        if l == 1:
            self.undoCmds.append(cmd)
            self.index = len(self.undoCmds) - 1
        else:
            self.macroStack[-1].append(cmd)
        self.clean = False
    def undo(self):
        assert(not self.macroStack)  # Can't undo in the middle of a macro!
        if self.index > 0:
            cmd = self.undoCmds[self.index]
            self.index = self.index - 1
            cmd.undo()
    def redo(self):
        assert(not macroStack)  # Can't redo in the middle of a macro
        if self.index < len(self.undoCmds) - 1:
            cmd = self.undoCmds[self.index + 1]
            self.index = self.index + 1
            cmd.redo()

class QColor(object):
    r = 0
    g = 0
    b = 0
    a = 0
    def __init__(self, *args):
        if len(args) == 1:
            assert(type(args[0]) == str)
            hvals = re.findall('[0-9a-fA-F]{2}', args[0])
            hvals = [int(hv, 16) for hv in hvals]
        elif len(args) == 3:
            hvals = list(args)
        elif len(args) == 4:
            hvals = list(args)
        else:
            hvals = []
        while len(hvals) < 3:
            hvals.append(0)
        if len(hvals) < 4:
            hvals.append(255)
        for hv in hvals:
            assert(0 <= hv <= 255)
        r, g, b, a = hvals

class QFont(object):
    dummy = True
    Bold = None
    def __init__(self, *args):
        pass

class QFontMetricsF(object):
    def __init__(self, *args):
        pass



class ActiveSliceItemController(object):
    def __init__(self, activeSliceItem, modelPart):
        self._activeSliceItem = activeSliceItem
        self._modelPart = modelPart
        self.connectSignals()

    def connectSignals(self):
        aSI = self._activeSliceItem
        mP = self._modelPart

        mP.partActiveSliceResizeSignal.connect(aSI.updateRectSlot)
        mP.partActiveSliceIndexSignal.connect(aSI.updateIndexSlot)
        mP.partStrandChangedSignal.connect(aSI.strandChangedSlot)
    # end def

    def disconnectSignals(self):
        aSI = self._activeSliceItem
        mP = self._modelPart

        mP.partActiveSliceResizeSignal.disconnect(aSI.updateRectSlot)
        mP.partActiveSliceIndexSignal.disconnect(aSI.updateIndexSlot)
        mP.partStrandChangedSignal.disconnect(aSI.strandChangedSlot)
    # end def
# end class



#!/usr/bin/env python
# encoding: utf-8

import cadnano2.util as util
from cadnano2.views import styles

from cadnano2.model.enum import StrandType

# import Qt stuff into the module namespace with PySide, PyQt4 independence
util.qtWrapImport('QtCore', globals(), ['QPointF', 'QRectF', 'Qt'])
util.qtWrapImport('QtGui', globals(), ['QBrush',
                                       'QFont',
                                       'QFontMetrics',
                                       'QPainterPath',
                                       'QPolygonF',
                                       'QPen',
                                       'QUndoCommand'])
util.qtWrapImport('QtWidgets', globals(), ['QGraphicsPathItem',
                                           'QGraphicsRectItem',
                                           'QGraphicsSimpleTextItem'])

# construct paths for breakpoint handles
def _hashMarkGen(path, p1, p2, p3):
    path.moveTo(p1)
    path.lineTo(p2)
    path.lineTo(p3)
# end

# create hash marks QPainterPaths only once
_ppRect = QRectF(0, 0, styles.PATH_BASE_WIDTH, styles.PATH_BASE_WIDTH)
_pathCenter = QPointF(styles.PATH_BASE_WIDTH / 2,\
                          styles.PATH_BASE_WIDTH / 2)
_pathUCenter = QPointF(styles.PATH_BASE_WIDTH / 2, 0)
_pathDCenter = QPointF(styles.PATH_BASE_WIDTH / 2, styles.PATH_BASE_WIDTH)
_ppathLU = QPainterPath()
_hashMarkGen(_ppathLU, _ppRect.bottomLeft(), _pathDCenter, _pathCenter)
_ppathRU = QPainterPath()
_hashMarkGen(_ppathRU, _ppRect.bottomRight(), _pathDCenter, _pathCenter)
_ppathRD = QPainterPath()
_hashMarkGen(_ppathRD, _ppRect.topRight(), _pathUCenter, _pathCenter)
_ppathLD = QPainterPath()
_hashMarkGen(_ppathLD, _ppRect.topLeft(), _pathUCenter, _pathCenter)

_scafpen = QPen(styles.pxi_scaf_stroke, styles.PATH_STRAND_STROKE_WIDTH)
_scafpen.setCapStyle(Qt.PenCapStyle.FlatCap)  # or Qt.RoundCap
_scafpen.setJoinStyle(Qt.PenJoinStyle.RoundJoin)
_stappen = QPen(styles.pxi_stap_stroke, styles.PATH_STRAND_STROKE_WIDTH)
_stappen.setCapStyle(Qt.PenCapStyle.FlatCap)  # or Qt.RoundCap
_stappen.setJoinStyle(Qt.PenJoinStyle.RoundJoin)
_disabpen = QPen(styles.pxi_disab_stroke, styles.PATH_STRAND_STROKE_WIDTH)
_disabpen.setCapStyle(Qt.PenCapStyle.FlatCap)
_disabpen.setJoinStyle(Qt.PenJoinStyle.RoundJoin)
_disabbrush = QBrush(styles.pxi_disab_stroke)  # For the helix number label
_enabbrush = QBrush(Qt.BrushStyle.SolidPattern)  # Also for the helix number label
_baseWidth = styles.PATH_BASE_WIDTH
_rect = QRectF(0, 0, styles.PATH_BASE_WIDTH, 1.2*styles.PATH_BASE_WIDTH)
_toHelixNumFont = styles.XOVER_LABEL_FONT
# precalculate the height of a number font.  Assumes a fixed font
# and that only numbers will be used for labels
_fm = QFontMetrics(_toHelixNumFont)

class PreXoverItem(QGraphicsPathItem):
    def __init__(self,  fromVirtualHelixItem, toVirtualHelixItem, index, strandType, isLowIdx):
        super(PreXoverItem, self).__init__(fromVirtualHelixItem)
        self._fromVHItem = fromVirtualHelixItem
        self._toVHItem = toVirtualHelixItem
        self._idx = index
        self._strandType = strandType
        # translate from Low to Left for the Path View
        self._isLowIndex = isLowIdx
        self._isActive = False
        self._pen = _scafpen if strandType == StrandType.Scaffold else _stappen
        isOnTop = fromVirtualHelixItem.isStrandTypeOnTop(strandType)

        bw = _baseWidth
        x = bw * index
        y = (-1.25 if isOnTop else 2.25) * bw
        self.setPos(x, y)

        num = toVirtualHelixItem.number()
        tBR = _fm.tightBoundingRect(str(num))
        halfLabelH = tBR.height()/2.0
        halfLabelW = tBR.width()/2.0

        labelX = bw/2.0 - halfLabelW #
        if num == 1:  # adjust for the number one
            labelX -= halfLabelW/2.0

        if isOnTop:
            labelY = -0.25*halfLabelH - .5
        else:
            labelY = 2*halfLabelH + .5

        self._label = QGraphicsSimpleTextItem(self)
        self._label.setPos(labelX, labelY)

        # create a bounding rect item to process click events
        # over a wide area
        self._clickArea = cA = QGraphicsRectItem(_rect, self)
        cA.mousePressEvent = self.mousePress
        yoffset = 0.2*bw if isOnTop else -0.4*bw
        cA.setPos(0, yoffset)
        cA.setPen(QPen(Qt.PenStyle.NoPen))

        self.updateStyle()
        self._updateLabel()
        self.setPainterPath()
    # end def

    ### DRAWING METHODS ###
    def remove(self):
        scene = self.scene()
        if scene:
            scene.removeItem(self._label)
            scene.removeItem(self._clickArea)
            scene.removeItem(self)
        self._label = None
        self._clickArea = None
        self._fromVHItem = None
        self._toVHItem = None
    # end def

    def setPainterPath(self):
        """
        Sets the PainterPath according to the index (low = Left, high = Right)
        and strand position (top = Up, bottom = Down).
        """
        pathLUT = (_ppathRD, _ppathRU, _ppathLD, _ppathLU)  # Lookup table
        vhi = self._fromVHItem
        st = self._strandType
        path = pathLUT[2*int(self._isLowIndex) + int(vhi.isStrandTypeOnTop(st))]
        self.setPath(path)
    # end def

    def updateStyle(self):
        """
        If a PreXover can be installed the pen is a bold color,
        otherwise the PreXover is drawn with a disabled or muted color
        """
        fromVH = self._fromVHItem.virtualHelix()
        toVH = self._toVHItem.virtualHelix()
        part = self._fromVHItem.part()
        pen = _disabpen
        self._labelBrush = _disabbrush
        if part.possibleXoverAt(fromVH, toVH, self._strandType, self._idx):
            pen = self._pen
            self._isActive = True
            self._labelBrush = _enabbrush
        self.setPen(pen)
    # end def

    def _updateLabel(self):
        lbl = self._label
        lbl.setBrush(self._labelBrush)
        lbl.setFont(_toHelixNumFont)
        lbl.setText( str(self._toVHItem.number() ) )
    # end def

    ### TOOL METHODS ###
    def selectToolMousePress(self, event):
        """removexover(fromStrand, fromIdx, toStrand, toIdx)"""
        pass
    # end def

    def mousePress(self, event):
        if event.button() != Qt.MouseButton.LeftButton:
            return QGraphicsPathItem.mousePressEvent(self, event)

        if event.modifiers() & Qt.KeyboardModifier.ShiftModifier:
            return  # ignore shift click, user is probably trying to merge

        if self._isActive:
            fromVH = self._fromVHItem.virtualHelix()
            toVH = self._toVHItem.virtualHelix()
            fromSS = fromVH.getStrandSetByType(self._strandType)
            toSS = toVH.getStrandSetByType(self._strandType)
            fromStrand = fromSS.getStrand(self._idx)
            toStrand = toSS.getStrand(self._idx)
            part = self._fromVHItem.part()
            # determine if we are a 5' or a 3' end
            if self.path() in [_ppathLU, _ppathRD]:  # 3' end of strand5p clicked
                strand5p = fromStrand
                strand3p = toStrand
            else:  # 5'
                strand5p = toStrand
                strand3p = fromStrand

            # Gotta clear selections when installing a prexover
            # otherwise parenting in screwed up
            self._fromVHItem.viewroot().clearStrandSelections()

            part.createXover(strand5p, self._idx, strand3p, self._idx)
        else:
            event.setAccepted(False)
    # end def
    


from cadnano2.views import styles
from cadnano2.model.enum import StrandType

import cadnano2.util as util
# import Qt stuff into the module namespace with PySide, PyQt4 independence
util.qtWrapImport('QtCore', globals(), ['QPointF', 'QRectF', 'Qt'])
util.qtWrapImport('QtGui', globals(), ['QBrush', 'QFont', 'QPen'])
util.qtWrapImport('QtWidgets', globals(), ['QGraphicsItem',
                                           'QGraphicsItemGroup',
                                           'QGraphicsObject'])


_bw = styles.PATH_BASE_WIDTH
_toolRect = QRectF(0, 0, _bw, _bw)  # protected not private
_rect = QRectF(-styles.PATH_BASE_HL_STROKE_WIDTH,
               -styles.PATH_BASE_HL_STROKE_WIDTH,
               _bw + 2*styles.PATH_BASE_HL_STROKE_WIDTH,
               _bw + 2*styles.PATH_BASE_HL_STROKE_WIDTH)
_pen = QPen(styles.redstroke, styles.PATH_BASE_HL_STROKE_WIDTH)
_brush = QBrush(Qt.BrushStyle.NoBrush)

# There's a bug where C++ will free orphaned graphics items out from
# under pyqt. To avoid this, "_mother" adopts orphaned graphics items.
_mother = QGraphicsItemGroup()


class AbstractPathTool(QGraphicsObject):
    """Abstract base class to be subclassed by all other pathview tools."""
    def __init__(self, controller, parent=None):
        super(AbstractPathTool, self).__init__(parent)
        self._controller = controller
        self._window = controller.window
        self._active = False
        self._lastLocation = None

    ######################## Drawing #######################################
    def paint(self, painter, option, widget=None):
        painter.setPen(self._pen)
        painter.setBrush(self._brush)
        painter.drawRect(self._toolRect)

    def boundingRect(self):
        return self._rect

    ######################### Positioning and Parenting ####################
    def hoverEnterVirtualHelixItem(self, virtualHelixItem, event):
        self.updateLocation(virtualHelixItem, virtualHelixItem.mapToScene(QPointF(event.position())))

    def hoverLeaveVirtualHelixItem(self, virtualHelixItem, event):
        self.updateLocation(None, virtualHelixItem.mapToScene(QPointF(event.position())))

    def hoverMoveVirtualHelixItem(self, virtualHelixItem, event, flag=None):
        self.updateLocation(virtualHelixItem, virtualHelixItem.mapToScene(QPointF(event.position())))

    def updateLocation(self, virtualHelixItem, scenePos, *varargs):
        """Takes care of caching the location so that a tool switch
        outside the context of an event will know where to
        position the new tool and snaps self's pos to the upper
        left hand corner of the base the user is mousing over"""
        if virtualHelixItem:
            if self.parentObject() != virtualHelixItem:
                self.setParentItem(virtualHelixItem)
            self._lastLocation = (virtualHelixItem, scenePos)
            posItem = virtualHelixItem.mapFromScene(scenePos)
            pos = self.helixPos(posItem)
            if pos != None:
                if pos != self.pos():
                    self.setPos(pos)
                self.update(self.boundingRect())
                if not self.isVisible():
                    self.show()
                    pass
        else:
            self._lastLocation = None
            if self.isVisible():
                self.hide()
            if self.parentItem() != _mother:
                self.setParentItem(_mother)

    def lastLocation(self):
        """A tuple (virtualHelixItem, QPoint) representing the last
        known location of the mouse for purposes of positioning
        the graphic of a new tool on switching tools (the tool
        will have updateLocation(*oldTool.lastLocation()) called
        on it)"""
        return self._lastLocation

    def setActive(self, willBeActive, oldTool=None):
        """
        Called by PathToolManager.setActiveTool when the tool becomes
        active. Used, for example, to show/hide tool-specific ui elements.
        """
        if self.isActive() and not willBeActive:
            self.setParentItem(_mother)
            self.hide()

    def isActive(self):
        """Returns isActive"""
        return self._active != _mother

    def widgetClicked(self):
        """Called every time a widget representing self gets clicked,
        not just when changing tools."""
        pass

    ####################### Coordinate Utilities ###########################
    def baseAtPoint(self, virtualHelixItem, pt):
        """Returns the (strandType, baseIdx) corresponding
        to pt in virtualHelixItem."""
        x, strandIdx = self.helixIndex(pt)
        vh = virtualHelixItem.virtualHelix()
        if vh.isEvenParity():
            strandType = (StrandType.Scaffold, StrandType.Staple)[util.clamp(strandIdx, 0, 1)]
        else:
            strandType = (StrandType.Staple, StrandType.Scaffold)[util.clamp(strandIdx, 0, 1)]
        return (strandType, x, strandIdx)

    def helixIndex(self, point):
        """
        Returns the (row, col) of the base which point
        lies within.
        point is in virtualHelixItem coordinates.
        """
        x = int(int(point.x()) / _bw)
        y = int(int(point.y()) / _bw)
        return (x, y)
    # end def

    def helixPos(self, point):
        """
        Snaps a point to the upper left corner of the base
        it is within.
        point is in virtualHelixItem coordinates
        """
        col = int(int(point.x()) / _bw)
        row = int(int(point.y()) / _bw)
        # Doesn't know numBases, can't check if point is too far right
        if col < 0 or row < 0 or row > 1:
            return None
        return QPointF(col*_bw, row*_bw)
    # end def
# end class


#!/usr/bin/env python
# encoding: utf-8

from cadnano2.cadnano import app
from .part import Part
from cadnano2.model.enum import LatticeType


class Crossovers:
    honeycombScafLow = [[1, 11], [8, 18], [4, 15]]
    honeycombScafHigh = [[2, 12], [9, 19], [5, 16]]
    honeycombStapLow = [[6], [13], [20]]
    honeycombStapHigh = [[7], [14], [0]]


root3 = 1.732051


class HoneycombPart(Part):
    _step = 21  # 32 in square
    _turnsPerStep = 2.0
    _helicalPitch = _step/_turnsPerStep
    _twistPerBase = 360/_helicalPitch  # degrees
    _twistOffset = 0 # degrees

    _activeBaseIndex = _step
    _subStepSize = _step / 3
    # Used in VirtualHelix::potentialCrossoverList
    _scafL = Crossovers.honeycombScafLow
    _scafH = Crossovers.honeycombScafHigh
    _stapL = Crossovers.honeycombStapLow
    _stapH = Crossovers.honeycombStapHigh

    def __init__(self, *args, **kwargs):
        super(HoneycombPart, self).__init__(self, *args, **kwargs)
        self._maxRow = kwargs.get('maxRow', app().prefs.honeycombRows)
        self._maxCol = kwargs.get('maxCol', app().prefs.honeycombCols)
        self._maxBase = int(kwargs.get('maxSteps', app().prefs.honeycombSteps) * self._step - 1)

    def crossSectionType(self):
        """Returns the cross-section type of the DNA part."""
        return LatticeType.Honeycomb

    def isEvenParity(self, row, column):
        return (row % 2) == (column % 2)
    # end def

    def isOddParity(self, row, column):
        return (row % 2) ^ (column % 2)
    # end def

    def getVirtualHelixNeighbors(self, virtualHelix):
        """
        returns the list of neighboring virtualHelices based on parity of an
        input virtualHelix

        If a potential neighbor doesn't exist, None is returned in it's place
        """
        neighbors = []
        vh = virtualHelix
        if vh == None:
            return neighbors

        # assign the method to a a local variable
        getVH = self.virtualHelixAtCoord
        # get the vh's row and column r,c
        (r, c) = vh.coord()

        if self.isEvenParity(r, c):
            neighbors.append(getVH((r,c+1)))  # p0 neighbor (p0 is a direction)
            neighbors.append(getVH((r-1,c)))  # p1 neighbor
            neighbors.append(getVH((r,c-1)))  # p2 neighbor
        else:
            neighbors.append(getVH((r,c-1)))  # p0 neighbor (p0 is a direction)
            neighbors.append(getVH((r+1,c)))  # p1 neighbor
            neighbors.append(getVH((r,c+1)))  # p2 neighbor
        # For indices of available directions, use range(0, len(neighbors))
        return neighbors  # Note: the order and presence of Nones is important
    # end def

    def latticeCoordToPositionXY(self, row, column, scaleFactor=1.0):
        """make sure self._radius is a float"""
        radius = self._radius
        x = column*radius*root3
        if self.isOddParity(row, column):   # odd parity
            y = row*radius*3 + radius
        else:                               # even parity
            y = row*radius*3
        return scaleFactor*x, scaleFactor*y
    # end def

    def positionToCoord(self, x, y, scaleFactor=1.0):
        radius = self._radius
        column = int(x/(radius*root3*scaleFactor) + 0.5)

        rowTemp = y/(radius*scaleFactor)
        if (rowTemp % 3) + 0.5 > 1.0:
            # odd parity
            row = int((rowTemp-1)/3 + 0.5)
        else:
            # even parity
            row = int(rowTemp/3 + 0.5)
    # end def

    ########################## Archiving / Unarchiving #########################
    def fillSimpleRep(self, sr):
        super(HoneycombPart, self).fillSimpleRep(sr)
        sr['.class'] = 'HoneycombPart'



"""
XML Test Runner for PyUnit
"""

# Written by Sebastian Rittau <srittau@jroger.in-berlin.de> and placed in
# the Public Domain. With contributions by Paolo Borelli and others.



__version__ = "0.1"

import os.path
import re
import sys
import time
import traceback
import unittest
from xml.sax.saxutils import escape

try:
    from io import StringIO
except ImportError:
    from io import StringIO


class _TestInfo(object):
    """Information about a particular test. Used by _XMLTestResult."""
    def __init__(self, test, time):
        (self._class, self._method) = test.id().rsplit(".", 1)
        self._time = time
        self._error = None
        self._failure = None

    @staticmethod
    def create_success(test, time):
        """Create a _TestInfo instance for a successful test."""
        return _TestInfo(test, time)

    @staticmethod
    def create_failure(test, time, failure):
        """Create a _TestInfo instance for a failed test."""
        info = _TestInfo(test, time)
        info._failure = failure
        return info

    @staticmethod
    def create_error(test, time, error):
        """Create a _TestInfo instance for an erroneous test."""
        info = _TestInfo(test, time)
        info._error = error
        return info

    def print_report(self, stream):
        """Print information about this test case in XML format to the
        supplied stream.

        """
        stream.write('  <testcase classname="%(class)s" name="%(method)s" time="%(time).4f">' % \
            {
                "class": self._class,
                "method": self._method,
                "time": self._time,
            })
        if self._failure is not None:
            self._print_error(stream, 'failure', self._failure)
        if self._error is not None:
            self._print_error(stream, 'error', self._error)
        stream.write('</testcase>\n')

    def _print_error(self, stream, tagname, error):
        """Print information from a failure or error to the supplied stream."""
        text = escape(str(error[1]))
        stream.write('\n')
        stream.write('    <%s type="%s">%s\n' \
            % (tagname, _clsname(error[0]), text))
        tb_stream = StringIO()
        traceback.print_tb(error[2], None, tb_stream)
        stream.write(escape(tb_stream.getvalue()))
        stream.write('    </%s>\n' % tagname)
        stream.write('  ')


def _clsname(cls):
    return cls.__module__ + "." + cls.__name__


class _XMLTestResult(unittest.TestResult):

    """A test result class that stores result as XML.

    Used by XMLTestRunner.

    """

    def __init__(self, classname):
        unittest.TestResult.__init__(self)
        self._test_name = classname
        self._start_time = None
        self._tests = []
        self._error = None
        self._failure = None

    def startTest(self, test):
        unittest.TestResult.startTest(self, test)
        self._error = None
        self._failure = None
        self._start_time = time.time()

    def stopTest(self, test):
        time_taken = time.time() - self._start_time
        unittest.TestResult.stopTest(self, test)
        if self._error:
            info = _TestInfo.create_error(test, time_taken, self._error)
        elif self._failure:
            info = _TestInfo.create_failure(test, time_taken, self._failure)
        else:
            info = _TestInfo.create_success(test, time_taken)
        self._tests.append(info)

    def addError(self, test, err):
        unittest.TestResult.addError(self, test, err)
        self._error = err

    def addFailure(self, test, err):
        unittest.TestResult.addFailure(self, test, err)
        self._failure = err

    def print_report(self, stream, time_taken, out, err):
        """Prints the XML report to the supplied stream.

        The time the tests took to perform as well as the captured standard
        output and standard error streams must be passed in.
        """
        stream.write('<testsuite errors="%(e)d" failures="%(f)d" ' % \
            {"e": len(self.errors), "f": len(self.failures)})
        stream.write('name="%(n)s" tests="%(t)d" time="%(time).3f">\n' % \
            {"n": self._test_name, "t": self.testsRun, "time": time_taken})
        for info in self._tests:
            info.print_report(stream)
        stream.write('  <system-out><![CDATA[%s]]></system-out>\n' % out)
        stream.write('  <system-err><![CDATA[%s]]></system-err>\n' % err)
        stream.write('</testsuite>\n')


class XMLTestRunner(object):

    """A test runner that stores results in XML format compatible with JUnit.

    XMLTestRunner(stream=None) -> XML test runner

    The XML file is written to the supplied stream. If stream is None, the
    results are stored in a file called TEST-<module>.<class>.xml in the
    current working directory (if not overridden with the path property),
    where <module> and <class> are the module and class name of the test class.

    """

    def __init__(self, stream=None):
        self._stream = stream
        self._path = "."

    def run(self, test):
        """Run the given test case or test suite."""
        class_ = test.__class__
        classname = class_.__module__ + "." + class_.__name__
        if self._stream == None:
            filename = "TEST-%s.xml" % classname
            stream = file(os.path.join(self._path, filename), "w")
        else:
            stream = self._stream

        stream.write('<?xml version="1.0" encoding="utf-8"?>\n')
        result = _XMLTestResult(classname)
        start_time = time.time()

        with _fake_std_streams():
            test(result)
            try:
                out_s = sys.stdout.getvalue()
            except AttributeError:
                out_s = ""
            try:
                err_s = sys.stderr.getvalue()
            except AttributeError:
                err_s = ""

        time_taken = time.time() - start_time
        result.print_report(stream, time_taken, out_s, err_s)
        if self._stream is None:
            stream.close()

        return result

    def _set_path(self, path):
        self._path = path

    path = property(lambda self: self._path, _set_path, None,
            """The path where the XML files are stored.

            This property is ignored when the XML file is written to a file
            stream.""")


class _fake_std_streams(object):

    def __enter__(self):
        self._orig_stdout = sys.stdout
        self._orig_stderr = sys.stderr
        sys.stdout = StringIO()
        sys.stderr = StringIO()

    def __exit__(self, exc_type, exc_val, exc_tb):
        sys.stdout = self._orig_stdout
        sys.stderr = self._orig_stderr


class XMLTestRunnerTest(unittest.TestCase):

    def setUp(self):
        self._stream = StringIO()

    def _try_test_run(self, test_class, expected):

        """Run the test suite against the supplied test class and compare the
        XML result against the expected XML string. Fail if the expected
        string doesn't match the actual string. All time attributes in the
        expected string should have the value "0.000". All error and failure
        messages are reduced to "Foobar".

        """

        runner = XMLTestRunner(self._stream)
        runner.run(unittest.makeSuite(test_class))

        got = self._stream.getvalue()
        # Replace all time="X.YYY" attributes by time="0.000" to enable a
        # simple string comparison.
        got = re.sub(r'time="\d+\.\d+"', 'time="0.000"', got)
        # Likewise, replace all failure and error messages by a simple "Foobar"
        # string.
        got = re.sub(r'(?s)<failure (.*?)>.*?</failure>', r'<failure \1>Foobar</failure>', got)
        got = re.sub(r'(?s)<error (.*?)>.*?</error>', r'<error \1>Foobar</error>', got)
        # And finally Python 3 compatibility.
        got = got.replace('type="builtins.', 'type="exceptions.')

        self.assertEqual(expected, got)

    def test_no_tests(self):
        """
        Regression test: Check whether a test run without any tests
        matches a previous run.
        """
        class TestTest(unittest.TestCase):
            pass
        self._try_test_run(TestTest, """<testsuite errors="0" failures="0" name="unittest.TestSuite" tests="0" time="0.000">
  <system-out><![CDATA[]]></system-out>
  <system-err><![CDATA[]]></system-err>
</testsuite>
""")

    def test_success(self):
        """
        Regression test: Check whether a test run with a successful test
        matches a previous run.
        """
        class TestTest(unittest.TestCase):
            def test_foo(self):
                pass
        self._try_test_run(TestTest, """<testsuite errors="0" failures="0" name="unittest.TestSuite" tests="1" time="0.000">
  <testcase classname="__main__.TestTest" name="test_foo" time="0.000"></testcase>
  <system-out><![CDATA[]]></system-out>
  <system-err><![CDATA[]]></system-err>
</testsuite>
""")

    def test_failure(self):
        """
        Regression test: Check whether a test run with a failing test
        matches a previous run.
        """
        class TestTest(unittest.TestCase):
            def test_foo(self):
                self.assertTrue(False)
        self._try_test_run(TestTest, """<testsuite errors="0" failures="1" name="unittest.TestSuite" tests="1" time="0.000">
  <testcase classname="__main__.TestTest" name="test_foo" time="0.000">
    <failure type="exceptions.AssertionError">Foobar</failure>
  </testcase>
  <system-out><![CDATA[]]></system-out>
  <system-err><![CDATA[]]></system-err>
</testsuite>
""")

    def test_error(self):
        """
        Regression test: Check whether a test run with a erroneous test
        matches a previous run.
        """
        class TestTest(unittest.TestCase):
            def test_foo(self):
                raise IndexError()
        self._try_test_run(TestTest, """<testsuite errors="1" failures="0" name="unittest.TestSuite" tests="1" time="0.000">
  <testcase classname="__main__.TestTest" name="test_foo" time="0.000">
    <error type="exceptions.IndexError">Foobar</error>
  </testcase>
  <system-out><![CDATA[]]></system-out>
  <system-err><![CDATA[]]></system-err>
</testsuite>
""")

    def test_stdout_capture(self):
        """
        Regression test: Check whether a test run with output to stdout
        matches a previous run.
        """
        class TestTest(unittest.TestCase):
            def test_foo(self):
                sys.stdout.write("Test\n")
        self._try_test_run(TestTest, """<testsuite errors="0" failures="0" name="unittest.TestSuite" tests="1" time="0.000">
  <testcase classname="__main__.TestTest" name="test_foo" time="0.000"></testcase>
  <system-out><![CDATA[Test
]]></system-out>
  <system-err><![CDATA[]]></system-err>
</testsuite>
""")

    def test_stderr_capture(self):
        """
        Regression test: Check whether a test run with output to stderr
        matches a previous run.
        """
        class TestTest(unittest.TestCase):
            def test_foo(self):
                sys.stderr.write("Test\n")
        self._try_test_run(TestTest, """<testsuite errors="0" failures="0" name="unittest.TestSuite" tests="1" time="0.000">
  <testcase classname="__main__.TestTest" name="test_foo" time="0.000"></testcase>
  <system-out><![CDATA[]]></system-out>
  <system-err><![CDATA[Test
]]></system-err>
</testsuite>
""")

    class NullStream(object):
        """A file-like object that discards everything written to it."""
        def write(self, buffer):
            pass

    def test_unittests_changing_stdout(self):
        """
        Check whether the XMLTestRunner recovers gracefully from unit tests
        that change stdout, but don't change it back properly.
        """
        class TestTest(unittest.TestCase):
            def test_foo(self):
                sys.stdout = XMLTestRunnerTest.NullStream()

        runner = XMLTestRunner(self._stream)
        runner.run(unittest.makeSuite(TestTest))

    def test_unittests_changing_stderr(self):
        """
        Check whether the XMLTestRunner recovers gracefully from unit tests
        that change stderr, but don't change it back properly.
        """
        class TestTest(unittest.TestCase):
            def test_foo(self):
                sys.stderr = XMLTestRunnerTest.NullStream()

        runner = XMLTestRunner(self._stream)
        runner.run(unittest.makeSuite(TestTest))


if __name__ == "__main__":
    unittest.main()



#!/usr/bin/env python
# encoding: utf-8

import cadnano2.util as util
util.qtWrapImport('QtGui', globals(), ['QGraphicsPathItem'])


class AbstractDecoratorItem(QGraphicsPathItem):
    def __init__(self, parent):
        """The parent should be a VirtualHelixItem."""
        if self.__class__ == AbstractDecoratorItem:
            e = "AbstractDecoratorItem should be subclassed."
            raise NotImplementedError(e)
        super(AbstractDecoratorItem, self).__init__(parent)
        self._strand = None
        self._oligo = None

    ### SIGNALS ###

    ### SLOTS ###
    def strandResizedSlot(self):
        """docstring for strandResizedSlot"""
        pass

    def sequenceAddedSlot(self, oligo):
        """docstring for sequenceAddedSlot"""
        pass

    def decoratorRemovedSlot(self, oligo):
        """docstring for sequenceClearedSlot"""
        pass

    ### METHODS ###

    ### COMMANDS ###



#!/usr/bin/env python
# encoding: utf-8

from math import floor
from cadnano2.controllers.itemcontrollers.strand.stranditemcontroller import StrandItemController
from .endpointitem import EndpointItem
from cadnano2.views import styles
from .xoveritem import XoverItem
from .decorators.insertionitem import InsertionItem

import cadnano2.views.pathview.pathselection as pathselection

import cadnano2.util as util
# import Qt stuff into the module namespace with PySide, PyQt4 independence
util.qtWrapImport('QtCore', globals(), ['pyqtSignal', 'QObject', 'Qt', 'QRectF'])
util.qtWrapImport('QtGui', globals(), ['QBrush',
                                       'QColor',
                                       'QFont',
                                       'QFontMetricsF',
                                       'QPen'])
util.qtWrapImport('QtWidgets', globals(), ['QGraphicsLineItem',
                                           'QGraphicsPathItem',
                                           'QGraphicsItem',
                                           'QGraphicsSimpleTextItem',
                                           'QGraphicsRectItem'])

_baseWidth = styles.PATH_BASE_WIDTH
_defaultRect = QRectF(0,0, _baseWidth, _baseWidth)
_noPen = QPen(Qt.PenStyle.NoPen)


class StrandItem(QGraphicsLineItem):
    _filterName = "strand"

    def __init__(self, modelStrand, virtualHelixItem, viewroot):
        """The parent should be a VirtualHelixItem."""
        super(StrandItem, self).__init__(virtualHelixItem)
        self._modelStrand = modelStrand
        self._virtualHelixItem = virtualHelixItem
        self._viewroot = viewroot
        self._activeTool = virtualHelixItem.activeTool()

        self._controller = StrandItemController(self, modelStrand)
        isDrawn5to3 = modelStrand.strandSet().isDrawn5to3()

        self._strandFilter = modelStrand.strandFilter()

        self._insertionItems = {}
        # caps
        self._lowCap = EndpointItem(self, 'low', isDrawn5to3)
        self._highCap = EndpointItem(self, 'high', isDrawn5to3)
        self._dualCap = EndpointItem(self, 'dual', isDrawn5to3)

        # orientation
        self._isDrawn5to3 = isDrawn5to3
        # self._isOnTop = virtualHelixItem.isStrandOnTop(modelStrand)
        # label
        self._seqLabel = QGraphicsSimpleTextItem(self)

        self.refreshInsertionItems(modelStrand)
        self._updateSequenceText()

        # create a larger click area rect to capture mouse events
        self._clickArea = cA = QGraphicsRectItem(_defaultRect, self)
        cA.mousePressEvent = self.mousePressEvent
        cA.setPen(_noPen)
        self.setAcceptHoverEvents(True)
        cA.setAcceptHoverEvents(True)
        cA.hoverMoveEvent = self.hoverMoveEvent

        self.setZValue(styles.ZSTRANDITEM)

        # xover comming from the 3p end
        self._xover3pEnd = XoverItem(self, virtualHelixItem)
        # initial refresh
        self._updateColor(modelStrand)
        self._updateAppearance(modelStrand)

        self.setZValue(styles.ZSTRANDITEM)
        self.setFlag(QGraphicsItem.GraphicsItemFlag.ItemIsSelectable)
    # end def

    ### SIGNALS ###

    ### SLOTS ###
    def strandResizedSlot(self, strand, indices):
        """docstring for strandResizedSlot"""
        lowMoved = self._lowCap.updatePosIfNecessary(self.idxs()[0])
        highMoved = self._highCap.updatePosIfNecessary(self.idxs()[1])
        group = self.group()
        self.tempReparent()  
        if lowMoved:
            self.updateLine(self._lowCap)
        if highMoved:
            self.updateLine(self._highCap)
        if strand.connection3p():
            self._xover3pEnd.update(strand)
        self.refreshInsertionItems(strand)
        self._updateSequenceText()
        if group:
            group.addToGroup(self)
    # end def

    def sequenceAddedSlot(self, oligo):
        """docstring for sequenceAddedSlot"""
        pass
    # end def

    def sequenceClearedSlot(self, oligo):
        """docstring for sequenceClearedSlot"""
        pass
    # end def

    def strandRemovedSlot(self, strand):
        # self._modelStrand = None
        self._controller.disconnectSignals()
        self._controller = None
        scene = self.scene()
        scene.removeItem(self._clickArea)
        scene.removeItem(self._highCap)
        scene.removeItem(self._lowCap)
        scene.removeItem(self._seqLabel)
        self._xover3pEnd.remove()
        self._xover3pEnd = None
        for insertionItem in self._insertionItems.values():
            insertionItem.remove()
        self._insertionItems = None
        self._clickArea = None
        self._highCap = None
        self._lowCap = None
        self._seqLabel = None
        self._modelStrand = None
        self._virtualHelixItem = None
        scene.removeItem(self)
    # end def

    def strandUpdateSlot(self, strand):
        """
        Slot for just updating connectivity and color, and endpoint showing
        """
        self._updateAppearance(strand)
    # end def

    def oligoAppearanceChangedSlot(self, oligo):
        strand = self._modelStrand
        self._updateColor(strand)
        if strand.connection3p():
            self._xover3pEnd._updateColor(strand)
        for insertion in self.insertionItems().values():
            insertion.updateItem()
    # end def

    def oligoSequenceAddedSlot(self, oligo):
        self._updateSequenceText()
    # end def

    def oligoSequenceClearedSlot(self, oligo):
        self._updateSequenceText()
    # end def

    def strandHasNewOligoSlot(self, strand):
        strand = self._modelStrand
        self._controller.reconnectOligoSignals()
        self._updateColor(strand)
        if strand.connection3p():
            self._xover3pEnd._updateColor(strand)
    # end def

    def strandInsertionAddedSlot(self, strand, insertion):
        self.insertionItems()[insertion.idx()] = \
                    InsertionItem(self._virtualHelixItem, strand, insertion)
    # end def
    def strandInsertionChangedSlot(self, strand, insertion):
        self.insertionItems()[insertion.idx()].updateItem()
    # end def

    def strandInsertionRemovedSlot(self, strand, index):
        instItem = self.insertionItems()[index]
        instItem.remove()
        del self.insertionItems()[index]
    # end def

    def strandDecoratorAddedSlot(self, strand, decorator):
        pass
    # end def
    def strandDecoratorChangedSlot(self, strand, decorator):
        pass
    # end def
    def strandDecoratorRemovedSlot(self, strand, index):
        pass
    # end def
    def strandModifierAddedSlot(self, strand, modifier):
        pass
    # end def
    def strandModifierChangedSlot(self, strand, modifier):
        pass
    # end def
    def strandModifierRemovedSlot(self, strand, index):
        pass
    # end def

    def selectedChangedSlot(self, strand, indices):
        self.selectIfRequired(self.partItem().document(), indices)
    # end def

    ### ACCESSORS ###
    def activeTool(self):
        return self._activeTool
    # end def

    def viewroot(self):
        return self._viewroot
    # end def

    def insertionItems(self):
        return self._insertionItems
    # end def

    def strand(self):
        return self._modelStrand
    # end def

    def strandFilter(self):
        return self._strandFilter
    # end def

    def idxs(self):
        return self._modelStrand.idxs()
    # end def

    def virtualHelixItem(self):
        return self._virtualHelixItem
    # end def

    def partItem(self):
        return self._virtualHelixItem.partItem()
    # end def

    def window(self):
        return self._virtualHelixItem.window()

    ### PUBLIC METHODS FOR DRAWING / LAYOUT ###
    def refreshInsertionItems(self, strand):
        iItems = self.insertionItems()
        iModel = strand.insertionsOnStrand()

        was_in_use = set(iItems)
        in_use = set()
        # add in the ones supposed to be there
        for insertion in iModel:
            idx = insertion.idx()
            in_use.add(idx)
            if idx in iItems:
                pass
            else:
                iItems[insertion.idx()] = \
                    InsertionItem(self._virtualHelixItem, strand, insertion)
        # end for

        # remove all in items
        not_in_use = was_in_use - in_use
        for index in not_in_use:
            iItems[index].remove()
            del iItems[index]
        # end for
    # end def

    def resetStrandItem(self, virtualHelixItem, isDrawn5to3):
        self.setParentItem(virtualHelixItem)
        self._virtualHelixItem = virtualHelixItem
        self.resetEndPointItems(isDrawn5to3)
    # end def

    def resetEndPointItems(self, isDrawn5to3):
        self._isDrawn5to3 = isDrawn5to3
        self._lowCap.resetEndPoint(isDrawn5to3)
        self._highCap.resetEndPoint(isDrawn5to3)
        self._dualCap.resetEndPoint(isDrawn5to3)
    # end def

    def updateLine(self, movedCap):
        # setup
        bw = _baseWidth
        cA = self._clickArea
        line = self.line()
        # set new line coords
        if movedCap == self._lowCap:
            p1 = line.p1()
            newX = self._lowCap.pos().x() + bw
            p1.setX(newX)
            line.setP1(p1)
            temp = cA.rect()
            temp.setLeft(newX)
            cA.setRect(temp)
        else:
            p2 = line.p2()
            newX = self._highCap.pos().x()
            p2.setX(newX)
            line.setP2(p2)
            temp = cA.rect()
            temp.setRight(newX)
            cA.setRect(temp)
        self.setLine(line)
    # end def

    ### PRIVATE SUPPORT METHODS ###
    def _updateAppearance(self, strand):
        """
        Prepare Strand for drawing, positions are relative to the VirtualHelixItem:
        1. Show or hide caps depending on L and R connectivity.
        2. Determine line coordinates.
        3. Apply paint styles.
        """
        # 0. Setup
        vhi = self._virtualHelixItem
        bw = _baseWidth
        halfBaseWidth = bw / 2.0
        lowIdx, highIdx = strand.lowIdx(), strand.highIdx()

        lUpperLeftX, lUpperLeftY = vhi.upperLeftCornerOfBase(lowIdx, strand)
        hUpperLeftX, hUpperLeftY = vhi.upperLeftCornerOfBase(highIdx, strand)
        lowCap = self._lowCap
        highCap = self._highCap
        dualCap = self._dualCap

        # 1. Cap visibilty
        lx = lUpperLeftX + bw  # draw from right edge of base
        lowCap.safeSetPos(lUpperLeftX, lUpperLeftY)
        if strand.connectionLow() != None:  # has low xover
            # if we are hiding it, we might as well make sure it is reparented to the StrandItem
            lowCap.restoreParent()
            lowCap.hide()
        else:  # has low cap
            if not lowCap.isVisible():
                lowCap.show()

        hx = hUpperLeftX  # draw to edge of base
        highCap.safeSetPos(hUpperLeftX, hUpperLeftY)
        if strand.connectionHigh() != None:  # has high xover
            # if we are hiding it, we might as well make sure it is reparented to the StrandItem
            highCap.restoreParent()
            highCap.hide()
        else:  # has high cap
            if not highCap.isVisible():
                highCap.show()

        # special case: single-base strand with no L or H connections,
        # (unconnected caps were made visible in previous block of code)
        if strand.length() == 1 and (lowCap.isVisible() and highCap.isVisible()):
            lowCap.hide()
            highCap.hide()
            dualCap.safeSetPos(lUpperLeftX, lUpperLeftY)
            dualCap.show()
        else:
            dualCap.hide()

        # 2. Xover drawing
        xo = self._xover3pEnd
        if strand.connection3p():
            xo.update(strand)
            xo.showIt()
        else:
            xo.restoreParent()
            xo.hideIt()

        # 3. Refresh insertionItems if necessary drawing
        self.refreshInsertionItems(strand)

        # 4. Line drawing
        hy = ly = lUpperLeftY + halfBaseWidth
        self.setLine(lx, ly, hx, hy)
        rectf = QRectF(lUpperLeftX+bw, lUpperLeftY, bw*(highIdx-lowIdx-1), bw)
        self._clickArea.setRect(rectf)
        self._updateHighlight(self.pen().color())
    # end def

    def _updateColor(self, strand):
        oligo = self._modelStrand.oligo()
        color = QColor(oligo.color())
        self._updateHighlight(color)
    # end def

    def _updateHighlight(self, color):
        """

        """
        oligo = self._modelStrand.oligo()
        penWidth = styles.PATH_STRAND_STROKE_WIDTH
        if oligo.shouldHighlight():
            color.setAlpha(128)
            penWidth = styles.PATH_STRAND_HIGHLIGHT_STROKE_WIDTH
        pen = QPen(color, penWidth)
        # pen.setCosmetic(True)
        brush = QBrush(color)
        pen.setCapStyle(Qt.PenCapStyle.FlatCap)
        self.setPen(pen)
        self._lowCap.updateHighlight(brush)
        self._highCap.updateHighlight(brush)
        self._dualCap.updateHighlight(brush)
    # end def

    def _updateSequenceText(self):
        """
        docstring for _updateSequenceText
        """
        bw = _baseWidth
        seqLbl = self._seqLabel
        strand = self.strand()

        seqTxt = strand.sequence()
        isDrawn3to5 = not self._isDrawn5to3
        textXCenteringOffset = styles.SEQUENCETEXTXCENTERINGOFFSET

        if seqTxt == '':
            seqLbl.hide()
            for iItem in self.insertionItems().values():
                iItem.hideSequence()
            return
        # end if

        strandSeqList = strand.getSequenceList()
        seqList = [x[1][0] for x in strandSeqList]
        insertSeqList = [(x[0], x[1][1]) for x in strandSeqList]

        iItems = self.insertionItems()
        for idx, seqTxt in insertSeqList:
            if seqTxt != '':
                iItems[idx].setSequence(seqTxt)

        if isDrawn3to5:
            seqList = seqList[::-1]

        seqTxt = ''.join(seqList)

        # seqLbl.setPen(QPen( Qt.PenStyle.NoPen))    # leave the Pen as None for unless required
        seqLbl.setBrush(QBrush(Qt.GlobalColor.black))
        seqLbl.setFont(styles.SEQUENCEFONT)

        # this will always draw from the 5 Prime end!
        seqX = 2*textXCenteringOffset + bw*strand.idx5Prime()
        seqY = styles.SEQUENCETEXTYCENTERINGOFFSET

        if isDrawn3to5:
            # offset it towards the bottom
            seqY += bw * .8
            # offset X by the reverse centering offset and the string length
            seqX += textXCenteringOffset
            # rotate the characters upside down this does not affect positioning
            # coordinate system, +Y is still Down, and +X is still Right
            seqLbl.setRotation(180)
            # draw the text and reverse the string to draw 5 prime to 3 prime
            # seqTxt = seqTxt[::-1]
        # end if
        seqLbl.setPos(seqX,seqY)
        seqLbl.setText(seqTxt)
        seqLbl.show()
    # end def

    ### EVENT HANDLERS ###
    def mousePressEvent(self, event):
        """
        Parses a mousePressEvent to extract strandSet and base index,
        forwarding them to approproate tool method as necessary.
        """
        activeToolStr = str(self._activeTool())
        self.scene().views()[0].addToPressList(self)
        idx = int(floor((event.pos().x()) / _baseWidth))
        self._virtualHelixItem.setActive(idx)
        toolMethodName =  activeToolStr + "MousePress"
        if hasattr(self, toolMethodName):
            getattr(self, toolMethodName)(event, idx)
        else:
            event.setAccepted(False)
    # end def

    def mouseMoveEvent(self, event):
        """
        Parses a mouseMoveEvent to extract strandSet and base index,
        forwarding them to approproate tool method as necessary.
        """
        toolMethodName = str(self._activeTool()) + "MouseMove"
        if hasattr(self, toolMethodName):
            idx = int(floor((event.pos().x()) / _baseWidth))
            getattr(self, toolMethodName)(idx)
    # end def

    def hoverLeaveEvent(self, event):
        self.partItem().updateStatusBar("")
    # end def

    def hoverMoveEvent(self, event):
        """
        Parses a mouseMoveEvent to extract strandSet and base index,
        forwarding them to approproate tool method as necessary.
        """
        vhiNum = self._virtualHelixItem.number()
        idx = int(floor((event.pos().x()) / _baseWidth))
        oligoLength = self._modelStrand.oligo().length()
        self.partItem().updateStatusBar("%d[%d]\tlength: %d" % (vhiNum, idx, oligoLength))
        toolMethodName = str(self._activeTool()) + "HoverMove"
        if hasattr(self, toolMethodName):
            getattr(self, toolMethodName)(idx)
    # end def

    def customMouseRelease(self, event):
        """
        Parses a mouseReleaseEvent to extract strandSet and base index,
        forwarding them to approproate tool method as necessary.
        """
        toolMethodName = str(self._activeTool()) + "MouseRelease"
        if hasattr(self, toolMethodName):
            idx = int(floor((event.pos().x()) / _baseWidth))
            getattr(self, toolMethodName)(idx)
    # end def

    ### TOOL METHODS ###
    def breakToolMousePress(self, event, idx):
        """Break the strand is possible."""
        mStrand = self._modelStrand
        mStrand.split(idx)
    # end def

    def breakToolHoverMove(self, idx):
        pass
        # mStrand = self._modelStrand
        # vhi = self._virtualHelixItem
        # breakTool = self._activeTool()
        # breakTool.updateHoverRect(vhi, mStrand, idx, show=True)
    # end def

    def selectToolMousePress(self, event, idx):
        event.setAccepted(False)
        currentFilterDict = self._viewroot.selectionFilterDict()
        if self.strandFilter() in currentFilterDict and self._filterName in currentFilterDict:
            selectionGroup = self._viewroot.strandItemSelectionGroup()
            mod = Qt.KeyboardModifier.MetaModifier
            if not (event.modifiers() & mod):
                selectionGroup.clearSelection(False)
            selectionGroup.setSelectionLock(selectionGroup)
            selectionGroup.pendToAdd(self)
            selectionGroup.pendToAdd(self._lowCap)
            selectionGroup.pendToAdd(self._highCap)
            selectionGroup.processPendingToAddList()
            event.setAccepted(True)
            return selectionGroup.mousePressEvent(event)
    # end def

    def eraseToolMousePress(self, event, idx):
        mStrand = self._modelStrand
        mStrand.strandSet().removeStrand(mStrand)
    # end def

    def insertionToolMousePress(self, event, idx):
        """Add an insert to the strand if possible."""
        mStrand = self._modelStrand
        mStrand.addInsertion(idx, 1)
    # end def

    def paintToolMousePress(self, event, idx):
        """Add an insert to the strand if possible."""
        mStrand = self._modelStrand
        if mStrand.isStaple():
            color = self.window().pathColorPanel.stapColorName()
        else:
            color = self.window().pathColorPanel.scafColorName()
        mStrand.oligo().applyColor(color)
    # end def

    def pencilToolHoverMove(self, idx):
        """Pencil the strand is possible."""
        mStrand = self._modelStrand
        vhi = self._virtualHelixItem
        activeTool = self._activeTool()

        if not activeTool.isFloatingXoverBegin():
            tempXover = activeTool.floatingXover()
            tempXover.updateFloatingFromStrandItem(vhi, mStrand, idx)
            # if mStrand.idx5Prime() == idx:
            #     tempXover.hide3prime()
            # elif mStrand.idx3Prime() != idx:
            #     tempXover.show3prime()
    # end def

    def pencilToolMousePress(self, event, idx):
        """Break the strand is possible."""
        mStrand = self._modelStrand
        vhi = self._virtualHelixItem
        partItem = vhi.partItem()
        activeTool = self._activeTool()

        if activeTool.isFloatingXoverBegin():
            # block xovers starting at a 5 prime end
            if mStrand.idx5Prime() == idx:
                return
            else:
                tempXover = activeTool.floatingXover()
                tempXover.updateBase(vhi, mStrand, idx)
                activeTool.setFloatingXoverBegin(False)
        else:
            activeTool.setFloatingXoverBegin(True)
            # install Xover
            activeTool.attemptToCreateXover(vhi, mStrand, idx)
    # end def

    def skipToolMousePress(self, event, idx):
        """Add an insert to the strand if possible."""
        mStrand = self._modelStrand
        mStrand.addInsertion(idx, -1)
    # end def

    def addSeqToolMousePress(self, event, idx):
        """
        Checks that a scaffold was clicked, and then calls apply sequence
        to the clicked strand via its oligo.
        """
        mStrand = self._modelStrand
        if mStrand.isScaffold():
            olgLen, seqLen = self._activeTool().applySequence(mStrand.oligo())
            if olgLen:
                msg = "Populated %d of %d scaffold bases." % (min(seqLen, olgLen), olgLen)
                if olgLen > seqLen:
                    d = olgLen - seqLen
                    msg = msg + " Warning: %d bases have no sequence." % d
                elif olgLen < seqLen:
                    d = seqLen - olgLen
                    msg = msg + " Warning: %d sequence bases unused." % d
                self.partItem().updateStatusBar(msg)
    # end def

    def restoreParent(self, pos=None):
        """
        Required to restore parenting and positioning in the partItem
        """
        # map the position
        # print "restoring parent si"
        self.tempReparent(pos)
        self.setSelectedColor(False)
        self.setSelected(False)
    # end def

    def tempReparent(self, pos=None):
        vhItem = self.virtualHelixItem()
        if pos == None:
            pos = self.scenePos()
        self.setParentItem(vhItem)
        tempP = vhItem.mapFromScene(pos)
        self.setPos(tempP)
    # end def

    def setSelectedColor(self, value):
        if value == True:
            color = QColor("#ff3333")
        else:
            oligo = self._modelStrand.oligo()
            color = QColor(oligo.color())
            if oligo.shouldHighlight():
                color.setAlpha(128)
        pen = self.pen()
        pen.setColor(color)
        self.setPen(pen)
    # end def

    def itemChange(self, change, value):
        # for selection changes test against QGraphicsItem.GraphicsItemChange.ItemSelectedChange
        # intercept the change instead of the has changed to enable features.
        if change == QGraphicsItem.GraphicsItemChange.ItemSelectedChange and self.scene():
            activeTool = self._activeTool()
            if str(activeTool) == "selectTool":
                viewroot = self._viewroot
                currentFilterDict = viewroot.selectionFilterDict()
                selectionGroup = viewroot.strandItemSelectionGroup()

                # only add if the selectionGroup is not locked out
                isNormalSelect = selectionGroup.isNormalSelect()
                if value == True and (self._filterName in currentFilterDict or not isNormalSelect):
                    if self._strandFilter in currentFilterDict:
                        if self.group() != selectionGroup:
                            self.setSelectedColor(True)
                            # This should always be the case, but...
                            if isNormalSelect:
                                selectionGroup.pendToAdd(self)
                                selectionGroup.setSelectionLock(selectionGroup)
                                selectionGroup.pendToAdd(self._lowCap)
                                selectionGroup.pendToAdd(self._highCap)
                            # this else will capture the error.  Basically, the
                            # strandItem should be member of the group before this
                            # ever gets fired
                            else:
                                selectionGroup.addToGroup(self)
                        return True
                    else:
                        return False
                # end if
                elif value == True:
                    # Don't select
                    return False
                # end elif
                else:
                    # Deselect
                    # print "Deselecting strand"
                    selectionGroup.pendToRemove(self)
                    self.setSelectedColor(False)
                    selectionGroup.pendToRemove(self._lowCap)
                    selectionGroup.pendToRemove(self._highCap)
                    return False
                # end else
            # end if
            elif str(activeTool) == "paintTool":
                viewroot = self._viewroot
                currentFilterDict = viewroot.selectionFilterDict()
                if self._strandFilter in currentFilterDict:
                    if not activeTool.isMacrod():
                        activeTool.setMacrod()
                    self.paintToolMousePress(None, None)
            # end elif
            return False
        # end if
        return QGraphicsItem.itemChange(self, change, value)
    # end def

    def selectIfRequired(self, document, indices):
        """
        Select self or xover item as necessary
        """
        strand5p = self._modelStrand
        con3p = strand5p.connection3p()
        selectionGroup = self._viewroot.strandItemSelectionGroup()
        # check this strand's xover
        idxL, idxH = indices
        if con3p:
            # perhaps change this to a direct call, but here are seeds of an 
            # indirect way of doing selection checks    
            if document.isModelStrandSelected(con3p) and document.isModelStrandSelected(strand5p):
                val3p = document.getSelectedStrandValue(con3p)
                # print "xover idx", indices
                test3p = val3p[0] if con3p.isDrawn5to3() else val3p[1]
                test5p = idxH if strand5p.isDrawn5to3() else idxL
                if test3p and test5p:
                    xoi = self._xover3pEnd
                    if not xoi.isSelected() or not xoi.group():
                        selectionGroup.setNormalSelect(False)
                        selectionGroup.addToGroup(xoi)
                        xoi.modelSelect(document)
                        selectionGroup.setNormalSelect(True)
                # end if
            # end if
        # end if
        # Now check the endpoints

        lowCap = self._lowCap
        if idxL == True:
            if not lowCap.isSelected() or not lowCap.group():
                selectionGroup.addToGroup(lowCap)
                lowCap.modelSelect(document)
        else:
            if lowCap.isSelected() or lowCap.group():
                lowCap.restoreParent()
        highCap = self._highCap
        if idxH == True:
            if not highCap.isSelected() or not highCap.group():
                selectionGroup.addToGroup(highCap)
                highCap.modelSelect(document)
        else:
            if highCap.isSelected() or highCap.group():
                highCap.restoreParent()

        # now check the strand itself
        if idxL == True and idxH == True:
            if not self.isSelected() or not self.group():
                selectionGroup.setNormalSelect(False)
                selectionGroup.addToGroup(self)
                self.modelSelect(document)
                selectionGroup.setNormalSelect(True)
        elif idxL == False and idxH == False:
            self.modelDeselect(document)
    # end def

    def modelDeselect(self, document):
        self.restoreParent()
        self._lowCap.modelDeselect(document)
        self._highCap.modelDeselect(document)
    # end def

    def modelSelect(self, document):
        self.setSelected(True)
        self.setSelectedColor(True)
    # end def

    def paint(self, painter, option, widget):
        painter.setPen(self.pen())
        painter.drawLine(self.line())
    # end def



from cadnano2.cadnano import app
from cadnano2.model.enum import LatticeType
from .part import Part


class Crossovers:
    squareScafLow = [[4, 26, 15], [18, 28, 7], [10, 20, 31], [2, 12, 23]]
    squareScafHigh = [[5, 27, 16], [19, 29, 8], [11, 21, 0], [3, 13, 24]]
    squareStapLow = [[31], [23], [15], [7]]
    squareStapHigh = [[0], [24], [16], [8]]


class SquarePart(Part):
    _step = 32  # 21 in honeycomb
    _subStepSize = _step / 4
    _turnsPerStep = 3.0
    _helicalPitch = _step/_turnsPerStep
    _twistPerBase = 360/_helicalPitch  # degrees
    _twistOffset = 180 + _twistPerBase/2  # degrees

    # Used in VirtualHelix::potentialCrossoverList
    _scafL = Crossovers.squareScafLow
    _scafH = Crossovers.squareScafHigh
    _stapL = Crossovers.squareStapLow
    _stapH = Crossovers.squareStapHigh

    def __init__(self, *args, **kwargs):
        super(SquarePart, self).__init__(self, *args, **kwargs)
        self._maxRow = kwargs.get('maxRow', app().prefs.squareRows)
        self._maxCol = kwargs.get('maxCol', app().prefs.squareCols)
        self._maxBase = int(kwargs.get('maxSteps', app().prefs.squareSteps) * self._step - 1)

    def crossSectionType(self):
        return LatticeType.Square
    # end def

    def isEvenParity(self, row, column):
        return (row % 2) == (column % 2)
    # end def

    def isOddParity(self, row, column):
        return (row % 2) ^ (column % 2)
    # end def

    def getVirtualHelixNeighbors(self, virtualHelix):
        """
        returns the list of neighboring virtualHelices based on parity of an
        input virtualHelix

        If a potential neighbor doesn't exist, None is returned in it's place
        """
        neighbors = []
        vh = virtualHelix
        if vh == None:
            return neighbors

        # assign the method to a a local variable
        getVH = self.virtualHelixAtCoord
        # get the vh's row and column r,c
        (r, c) = vh.coord()

        if self.isEvenParity(r, c):
            neighbors.append(getVH((r,c+1)))  # p0 neighbor (p0 is a direction)
            neighbors.append(getVH((r+1,c)))  # p1 neighbor
            neighbors.append(getVH((r,c-1)))  # p2 neighbor
            neighbors.append(getVH((r-1,c)))  # p2 neighbor
        else:
            neighbors.append(getVH((r,c-1)))  # p0 neighbor (p0 is a direction)
            neighbors.append(getVH((r-1,c)))  # p1 neighbor
            neighbors.append(getVH((r,c+1)))  # p2 neighbor
            neighbors.append(getVH((r+1,c)))  # p3 neighbor
        # For indices of available directions, use range(0, len(neighbors))
        return neighbors  # Note: the order and presence of Nones is important
    # end def

    def latticeCoordToPositionXY(self, row, column, scaleFactor=1.0):
        """
        make sure self._radius is a float
        """
        radius = self._radius
        y = row*2*radius
        x = column*2*radius
        return scaleFactor*x, scaleFactor*y
    # end def

    def positionToCoord(self, x, y, scaleFactor=1.0):
        """
        """
        radius = self._radius
        row = int(y/(2*radius*scaleFactor) + 0.5)
        column = int(x/(2*radius*scaleFactor) + 0.5)
        return row, column
    # end def

    def fillSimpleRep(self, sr):
        super(SquarePart, self).fillSimpleRep(sr)
        sr['.class'] = 'SquarePart'



# Copyright 2011 Autodesk, Inc.  All rights reserved.
#
# The MIT License
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
# http://www.opensource.org/licenses/mit-license.php

"""
partitem.py
Created by Simon Breslav on 2011-07-21
"""

from string import *
import math
import random
import os
import sys
from cadnano2.views import styles
from cadnano2.model.enum import LatticeType
from cadnano2.model.enum import StrandType
from cadnano2.model.virtualhelix import VirtualHelix

import maya.OpenMayaUI as mui
import maya.OpenMaya as mo
import maya.cmds as cmds
import cadnano2.util as util

from controllers.mayacontrollers.mayaObjectManager import Mom
from cadnano2.controllers.itemcontrollers.partitemcontroller import PartItemController
from .virtualhelixitem import VirtualHelixItem

# import Qt stuff into the module namespace with PySide, PyQt4 independence
util.qtWrapImport('QtCore', globals(), ['pyqtSignal', 'pyqtSlot', 'QObject'])


class PartItem(object):
    """
    PartItem stores VirtualHelixItems for a given DNA Part model
    """
    def __init__(self, modelPart, parent=None):
        """
        Loads in all the Maya plugins (Nodes) needed for visualizing
        the model (XXX [SB] -this code should probably go somewhere else).
        Initiates some private variables that are constant across all
        strands (XXX [SB] -this code should also probably go somewhere else).
        Sets up PartItemController that is used to setup all the
        slots and signals between strand model and this PartItem.
        """
        self._parentItem = parent
        pluginPath = os.path.join(os.environ['CADNANO_PATH'],
                                  "views",
                                  "solidview")
        hchPath = os.path.join(pluginPath, "halfcylinderhelixnode.py")
        smiPath = os.path.join(pluginPath, "predecoratornode.py")

        if(not cmds.pluginInfo(hchPath, query=True, loaded=True)):
            cmds.loadPlugin(hchPath)

        if(not cmds.pluginInfo(smiPath, query=True, loaded=True)):
            cmds.loadPlugin(smiPath)

        if(not cmds.pluginInfo(hchPath, query=True, loaded=True)):
            print("HalfCylinderHelixNode failed to load")
            return

        #print "maya PartItem created"
        self._type = modelPart.crossSectionType()

        #self.mayaScale = 1.0
        #later updates using basecount from the VH
        # XXX [SB] - need to ask CandNano for rows and cols...
        # top left cornder of maya 3d scene X Y Z
        self.mayaOrigin = (-15 * 2.25, 16 * 2.25, 0.0)
        self.helixRadius = 1.125  # diamiter is 2.25nm
        self._virtualHelixItems = {}

        self._part = modelPart
        self._controller = PartItemController(self, modelPart)

        self.modifyState = False
    # end def

    def parentItem(self):
        """Return parent item, which is SolidRootItem in this case"""
        return self._parentItem
    # end def

    def setModifyState(self, val):
        """Change Modify state for all the strands in this PartItem"""
        self.modifyState = val
        self.updateModifyState()
    # end def

    def updateModifyState(self):
        """Update Modify state for all the strands in this PartItem"""
        for sh in self._virtualHelixItems:
            sh.setModifyState(self.modifyState)
            sh.updateDecorators()
    # end def

    def isInModifyState(self):
        """Accessor for Modify State"""
        return self.modifyState
    # end def

    def type(self):
        """Accessor for Cross Section Type, (Honeycomb vs. Square)"""
        return self._type
    # end def

    def part(self):
        """Accessor for the part model"""
        return self._part
    # end def

    def setPart(self, p):
        """set part model"""
        self._part = p
    # end def

    ### SLOTS ###
    def partDimensionsChangedSlot(self, part):
        """
        Receives notification from the model when a dimentions of the part
        changes. Needs to change an attribute of every Maya Helix Nodes, so
        that they are all are aligned corectly with new strands that are
        created.
        """
        mom = Mom()
        for vh in self._virtualHelixItems:
            for mID in vh.StrandIDs():
                cylinderName = "%s%s" % (mom.helixNodeName, mID)
                totalNumBases = self._part.maxBaseIdx()
                cmds.setAttr("%s.totalBases" % cylinderName,
                                                int(totalNumBases))
            vh.updateDecorators()
    # end def

    def partParentChangedSlot(self, sender, part):
        """partParentChangedSlot - empty"""
        pass
    # end def

    def partRemovedSlot(self, sender, part):
        """clears out private variables and disconnects signals"""
        # print "solidview.PartItem.partRemovedSlot"
        self._virtualHelixItems = None
        self._parentItem.removePartItem(self)
        self._parentItem = None
        self._part = None
        self._controller.disconnectSignals()
        self._controller = None
    # end def

    def partPreDecoratorSelectedSlot(self, sender, row, col, baseIdx):
        """partPreDecoratorSelectedSlot - empty"""
        pass
    # end def

    def partVirtualHelixAddedSlot(self, sender, virtualHelix):
        """Receives notification when new VitualHelix is added"""
        sh = self.createNewVirtualHelixItem(virtualHelix)
        sh.setModifyState(self.modifyState)
    # end def

    @pyqtSlot(tuple)
    def partVirtualHelixRenumberedSlot(self, sender, coord):
        """partVirtualHelixRenumberedSlot - empty"""
        pass
    # end def

    @pyqtSlot(tuple)
    def partVirtualHelixResizedSlot(self, sender, coord):
        """partVirtualHelixResizedSlot - empty"""
        pass
    # end def

    @pyqtSlot(list)
    def partVirtualHelicesReorderedSlot(self, sender, orderedCoordList):
        """partVirtualHelicesReorderedSlot - empty"""
        pass
    # end def

    def updatePreXoverItemsSlot(self, sender, virtualHelix):
        """updatePreXoverItemsSlot - empty"""
        pass
    # end def

    ### METHODS ###
    def cadnanoToMayaCoords(self, row, col):
        """Converts cadnano row and col to Maya coordinates"""
        x, y = self.part().latticeCoordToPositionXY(row, col)
        return x + self.mayaOrigin[0], self.mayaOrigin[1] - y
    # end def

    def createNewVirtualHelixItem(self, virtualHelix):
        """Create a new Virtual Helix """
        coords = virtualHelix.coord()
        #print "solidview.PartItem.createNewVirtualHelixItem: %d %d" % \
        #                                                (coords[0], coords[1])
        # figure out Maya Coordinates
        x, y = self.cadnanoToMayaCoords(coords[0], coords[1])
        #print virtualHelix
        newHelix = VirtualHelixItem(self, virtualHelix, x, y)
        self._virtualHelixItems[newHelix] = True
        return newHelix
    # end def

    def removeVirtualHelixItem(self, vhelixItem):
        """Remove a new Virtual Helix """
        del self._virtualHelixItems[vhelixItem]
    # end def
# end class



# Form implementation generated from reading ui file 'dialogs/latticetype.ui'
#
# Created by: PyQt6 UI code generator 6.1.1
#
# WARNING: Any manual changes made to this file will be lost when pyuic6 is
# run again.  Do not edit this file unless you know what you are doing.


from PyQt6 import QtCore, QtGui, QtWidgets


class Ui_LatticeType(object):
    def setupUi(self, LatticeType):
        LatticeType.setObjectName("LatticeType")
        LatticeType.resize(215, 80)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Policy.Preferred, QtWidgets.QSizePolicy.Policy.Preferred)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(LatticeType.sizePolicy().hasHeightForWidth())
        LatticeType.setSizePolicy(sizePolicy)
        LatticeType.setSizeGripEnabled(False)
        self.verticalLayout = QtWidgets.QVBoxLayout(LatticeType)
        self.verticalLayout.setObjectName("verticalLayout")
        self.label = QtWidgets.QLabel(LatticeType)
        self.label.setObjectName("label")
        self.verticalLayout.addWidget(self.label)
        self.buttonBox = QtWidgets.QDialogButtonBox(LatticeType)
        self.buttonBox.setOrientation(QtCore.Qt.Orientation.Horizontal)
        self.buttonBox.setStandardButtons(QtWidgets.QDialogButtonBox.StandardButton.No|QtWidgets.QDialogButtonBox.StandardButton.Yes)
        self.buttonBox.setCenterButtons(True)
        self.buttonBox.setObjectName("buttonBox")
        self.verticalLayout.addWidget(self.buttonBox)

        self.retranslateUi(LatticeType)
        self.buttonBox.accepted.connect(LatticeType.accept)
        self.buttonBox.rejected.connect(LatticeType.reject)
        QtCore.QMetaObject.connectSlotsByName(LatticeType)

    def retranslateUi(self, LatticeType):
        _translate = QtCore.QCoreApplication.translate
        LatticeType.setWindowTitle(_translate("LatticeType", "Import"))
        self.label.setText(_translate("LatticeType", "Is this a square lattice design?"))



import time
import unittest
# Needs fixing - SD
# from PyQt4.QtCore import *
# from PyQt4.QtGui import *

main = unittest.main


class GUITestCase(unittest.TestCase):
    __qAppInitialized = False

    def __init__(self, *args, **kwargs):
        """
        We have to be careful on whether we do initialize a QApplication,
        so, we have to store a strong reference in an instance and store
        whether we did this in the class itself. When initializing again,
        we should not create another QApplication.

        (The unittest framework creates all classes and then calls setUp-
        tearDown on each one)
        """
        unittest.TestCase.__init__(self, *args, **kwargs)

        if not GUITestCase.__qAppInitialized:
            GUITestCase.__qAppInitialized = True
            self._qApplicationFirstReference = QApplication([])
        self._app = qApp
        self._app.processEvents()
        self._wait = 0

    def tearDown(self):
        self._test_widget.close()

    def setWidget(self, widget, show=True, wait=None):
        """
        Must be called in the setUp() method, giving the test widget.

        @param show: If show() should be called on the GUI. Set to False if
        you don't want to see the GUI running.

        @param wait: How long to wait between events, in seconds.
        """
        self._test_widget = widget
        self._wait = wait
        if show:
            self._test_widget.show()
        self.processEvents()

    # button flags
    NOBUTTON = Qt.MouseButton.NoButton
    LEFT = Qt.MouseButton.LeftButton
    RIGHT = Qt.MouseButton.RightButton
    MIDDLE = Qt.MouseButton.MidButton
    SHIFT = Qt.KeyboardModifier.ShiftModifier
    CONTROL = Qt.KeyboardModifier.ControlModifier
    ALT = Qt.KeyboardModifier.AltModifier

    ############################ Mouse events ############################
    def mousePress(self, widget, button=None, position=None, state=None, modifiers=None, qgraphicsscene=None):
        """Sends a press event for the given widget.
        @param button: the pressed mouse button. Can be LEFT, RIGHT or MIDDLE.
            If not given, LEFT is assumed.
        @param position: a QPoint or a pair, indicating the position in widget
            coordinates where the button was pressed. If not given, the center
            of the widget is used.
        @param state: secondary keys. optional, can be SHIFT, CONTROL, ALT.
        """
        self._mouseEvent(QEvent.Type.MouseButtonPress, widget, button, position, state, modifiers, qgraphicsscene)

    def mouseRelease(self, widget, button=None, position=None, state=None, modifiers=None, qgraphicsscene=None):
        """Sends a mouse release event for the given widget.

        @see: mousePress for the meaning of the arguments.
        """
        self._mouseEvent(QEvent.Type.MouseButtonRelease, widget, button, position, state, modifiers, qgraphicsscene)

    def mouseMove(self, widget, position=None, state=None, modifiers=None, qgraphicsscene=None):
        """Sends a mouse move event for the given widget.

        @see: mousePress for the meaning of the arguments.
        """
        state = self.LEFT
        self._mouseEvent(QEvent.Type.MouseMove, widget, self.NOBUTTON, position, state, modifiers, qgraphicsscene)

    def mouseDrag(self, widget, pressOn, releaseOn, button=None, state=None):
        """
        Makes a drag with the mouse.

        @pressOn: this is the position where the mouse is pressed.
        @releaseOn: this is the position where the mouse is released.
        """
        if state is None:
            state = self.LEFT
        self.mousePress(widget, button, pressOn, state)
        self.mouseMove(widget, releaseOn, state)
        self.mouseRelease(widget, button, releaseOn, state)

    def click(self, widget, button=None, position=None, state=None, modifiers=None, qgraphicsscene=None):
        """
        Acts as if the given widget was clicked. Equivalent to send a
        mousePress followed by a mouseRelease.

        @see: mousePress for the meaning of the arguments.
        """
        widget.setFocus()
        self.mousePress(widget, button, position, state, modifiers, qgraphicsscene)
        self.mouseRelease(widget, button, position, state, modifiers, qgraphicsscene)

    def doubleClick(self, widget, button=None, position=None, state=None):
        """Sends a double-click event to the given widget.

        @see: mousePress for the meaning of the arguments.
        """
        widget.setFocus()
        self._mouseEvent(QEvent.MouseButtonDblClick, widget, button, position, state)

    ########################## Keyboard events ############################
    def keyPress(self, widget, key, state=None):
        """Sends a key press event to the given widget.
        @param key: a qt.Qt.Key.Key_* constant or a one-char string.
        @param state: secondary keys. optional, can be SHIFT, CONTROL, ALT.
        """
        self._keyEvent(QEvent.KeyPress, widget, key, state)

    def keyRelease(self, widget, key, state=None):
        """Sends a key release event to the given widget.
        @see: keyPress for the meaning of the arguments.
        """
        self._keyEvent(QEvent.KeyRelease, widget, key, state)

    def type(self, widget, key, state=None):
        """Acts as if a key was typed in the given widget. Equivalent to a
        keyPress followed by a keyRelease.
        @see: keyPress for the meaning of the arguments.
        """
        self.keyPress(widget, key, state)
        self.keyRelease(widget, key, state)

    def typeText(self, widget, text):
        """Types the text over the given widget."""
        for char in text:
            self.type(widget, char)

    ########################## Miscellaneous ############################
    def debugHere(self):
        """Stops the executing of the test at the caller's position, returning
        control to qt's main loop. Useful to debug the tests.
        """
        self._app.setActiveWindow(self._test_widget)
        self._app.exec_()

    def processEvents(self):
        self._app.processEvents()
        if self._wait:
            time.sleep(self._wait)

    ############################ Private Methods ############################
    def _getWidgetCenter(self, widget):
        if getattr(widget, "boundingRect", False):
            # Can't pass a QPointF to QPoint constructor, so we let
            # python cast x and y to ints instead...
            return widget.boundingRect().center().toPoint()
        else:
            return QPoint(widget.width()/2, widget.height()/2)


    def _mouseEvent(self, event_type, widget, button, position, state, modifiers, qgraphicsscene=None):
        """Sends a MouseEvent with the given event_type.
        """
        if button is None:
            button = self.LEFT
        if position is None:
            position = self._getWidgetCenter(widget)
        if not isinstance(position, QPoint):
            position = QPoint(*position)  # assume position is a tuple
        if state is None:
            state = self.NOBUTTON
        if modifiers is None:
            modifiers = Qt.NoModifier
        event = QMouseEvent(event_type, position, button, state, modifiers)
        if qgraphicsscene:
            result = qgraphicsscene.sendEvent(widget, event)
        else:
            self._app.postEvent(widget, event)
        self.processEvents()

    def _keyEvent(self, event_type, widget, key, state=None):
        """Sends a key event.
        @param key: must be either a qt.Qt.Key.Key_* constant or a one-char string
        in string.printable.
        """
        if state is None:
            if isinstance(key, str) and key.isupper():
                state = Qt.KeyboardModifier.ShiftModifier
            else:
                state = self.NOBUTTON
        new_key, ascii = self._convertKey(key)
        string = _QT_TO_STR.get(new_key, '')
        if isinstance(key, str) and key.isupper():
            string = string.upper()
        if state is None:
            state = Qt.NoModifier
        event = Qt.QKeyEvent(event_type, new_key, state, string)
        self._app.postEvent(widget, event)
        self.processEvents()

    def _convertKey(self, key):
        """
        Handles the given key for a KeyEvent. Returns (key, ascii), where
        key is one of the qt.Qt.Key.Key_* constants. If key is a string, it is
        converted to one of the qt.Qt.Key.Key_* constants.
        """
        if isinstance(key, str):
            if key.lower() in _STR_TO_QT:
                ascii = ord(key)
                key = _STR_TO_QT[key.lower()]
                return key, ascii
            else:
                raise ValueError('Invalid key: %s' % key)
        else:
            ascii = ord(_QT_TO_STR.get(key, '\0'))
            return key, ascii

    def executeOnElapsed(self, callback, msecs):
        """
        This function schedules the callback to be executed when 'millis' has
        elapsed
        @param callback: this is the callable that should be executed.
        @param msecs: milliseconds to call callback.
        """
        QTimer.singleShot(msecs, callback)

_toElapse = []  # List of objects that we don't want garbage collected

KEY_RETURN = '\13'

# constant tables
constants = [
    (Qt.Key.Key_Escape, ''),
    (Qt.Key.Key_Tab, '\t'),
    (Qt.Key.Key_Backspace, '\b'),
    (Qt.Key.Key_Return, KEY_RETURN),
    (Qt.Key.Key_Enter, KEY_RETURN),
    (Qt.Key.Key_Space, ' '),
    (Qt.Key.Key_Exclam, '!'),
    (Qt.Key.Key_QuoteDbl, '"'),
    (Qt.Key.Key_NumberSign, '#'),
    (Qt.Key.Key_Dollar, '$'),
    (Qt.Key.Key_Percent, '%'),
    (Qt.Key.Key_Ampersand, '^'),
    (Qt.Key.Key_Apostrophe, '&'),
    (Qt.Key.Key_ParenLeft, '('),
    (Qt.Key.Key_ParenRight, ')'),
    (Qt.Key.Key_Asterisk, '*'),
    (Qt.Key.Key_Plus, '+'),
    (Qt.Key.Key_Comma, ','),
    (Qt.Key.Key_Minus, '-'),
    (Qt.Key.Key_Period, '.'),
    (Qt.Key.Key_Slash, '/'),
    (Qt.Key.Key_0, '0'),
    (Qt.Key.Key_1, '1'),
    (Qt.Key.Key_2, '2'),
    (Qt.Key.Key_3, '3'),
    (Qt.Key.Key_4, '4'),
    (Qt.Key.Key_5, '5'),
    (Qt.Key.Key_6, '6'),
    (Qt.Key.Key_7, '7'),
    (Qt.Key.Key_8, '8'),
    (Qt.Key.Key_9, '9'),
    (Qt.Key.Key_Colon, ':'),
    (Qt.Key.Key_Semicolon, ';'),
    (Qt.Key.Key_Less, '<'),
    (Qt.Key.Key_Equal, '='),
    (Qt.Key.Key_Greater, '>'),
    (Qt.Key.Key_Question, '?'),
    (Qt.Key.Key_At, '@'),
    (Qt.Key.Key_A, 'a'),
    (Qt.Key.Key_B, 'b'),
    (Qt.Key.Key_C, 'c'),
    (Qt.Key.Key_D, 'd'),
    (Qt.Key.Key_E, 'e'),
    (Qt.Key.Key_F, 'f'),
    (Qt.Key.Key_G, 'g'),
    (Qt.Key.Key_H, 'h'),
    (Qt.Key.Key_I, 'i'),
    (Qt.Key.Key_J, 'j'),
    (Qt.Key.Key_K, 'k'),
    (Qt.Key.Key_L, 'l'),
    (Qt.Key.Key_M, 'm'),
    (Qt.Key.Key_N, 'n'),
    (Qt.Key.Key_O, 'o'),
    (Qt.Key.Key_P, 'p'),
    (Qt.Key.Key_Q, 'q'),
    (Qt.Key.Key_R, 'r'),
    (Qt.Key.Key_S, 's'),
    (Qt.Key.Key_T, 't'),
    (Qt.Key.Key_U, 'u'),
    (Qt.Key.Key_V, 'v'),
    (Qt.Key.Key_W, 'w'),
    (Qt.Key.Key_X, 'x'),
    (Qt.Key.Key_Y, 'y'),
    (Qt.Key.Key_Z, 'z'),
    (Qt.Key.Key_BracketLeft, '['),
    (Qt.Key.Key_Backslash, '\\'),
    (Qt.Key.Key_BracketRight, ']'),
    (Qt.Key.Key_Underscore, '_'),
    (Qt.Key.Key_BraceLeft, '{'),
    (Qt.Key.Key_Bar, '|'),
    (Qt.Key.Key_BraceRight, '}'),
]

_QT_TO_STR = dict(constants)
_STR_TO_QT = dict([(y, x) for x, y in constants])
del constants



class VirtualHelixItemController():
    def __init__(self, virtualHelixItem, modelVirtualHelix):
        self._virtualHelixItem = virtualHelixItem
        self._modelVirtualHelix = modelVirtualHelix
        self.connectSignals()
    # end def

    def connectSignals(self):
        vhItem = self._virtualHelixItem
        mvh = self._modelVirtualHelix

        mvh.virtualHelixNumberChangedSignal.connect(vhItem.virtualHelixNumberChangedSlot)
        mvh.virtualHelixRemovedSignal.connect(vhItem.virtualHelixRemovedSlot)

        for strandSet in mvh.getStrandSets():
            strandSet.strandsetStrandAddedSignal.connect(vhItem.strandAddedSlot)
            # strandSet.decoratorAddedSignal.connect(vhItem.decoratorAddedSlot)
    # end def

    def disconnectSignals(self):
        vhItem = self._virtualHelixItem
        mvh = self._modelVirtualHelix

        mvh.virtualHelixNumberChangedSignal.disconnect(vhItem.virtualHelixNumberChangedSlot)
        mvh.virtualHelixRemovedSignal.disconnect(vhItem.virtualHelixRemovedSlot)

        for strandSet in mvh.getStrandSets():
            strandSet.strandsetStrandAddedSignal.disconnect(vhItem.strandAddedSlot)
            # strandSet.decoratorAddedSignal.disconnect(vhItem.decoratorAddedSlot)


"""
runtests.py
Helper file to run tests with 1 command and output the results to an XML file.
Usage: From the main cadnano folder: python -m tests.runtests
Created by Flo Mazzoldi on 2011-06-15.

My TextMate Script:

cd "$TM_PROJECT_DIRECTORY"
CADNANO_RUN_PLAINTEXT_TESTS=YES CADNANO_IGNORE_ENV_VARS_EXCEPT_FOR_ME=YES python tests/runall.py | pre
"""

import glob
import os
import unittest
from .xmlrunner import XMLTestRunner
# from unittests import UnitTests
# from modeltests import ModelTests
from .functionaltests import FunctionalTests
# from recordedtests.template import RecordedTests

def main(useXMLRunner=True):
    # load hard-coded tests
    # unitsuite = unittest.makeSuite(UnitTests)
    # modelsuite = unittest.makeSuite(ModelTests)
    funsuite = unittest.makeSuite(FunctionalTests)

    # combine and run tests
    # alltests = unittest.TestSuite([unitsuite, modelsuite, funsuite])
    alltests = unittest.TestSuite([funsuite])
    if useXMLRunner:
        stream = file("testresults.xml", "w")
        runner = XMLTestRunner(stream)
        result = runner.run(alltests)
        stream.close()
    else:
        runner = unittest.TextTestRunner(verbosity=2)
        result = runner.run(alltests)
    return result


if __name__ == "__main__":
    textRunner = os.environ.get('CADNANO_RUN_PLAINTEXT_TESTS', None) == "YES"
    main(not textRunner)



import sys, imp
import os.path
from glob import glob
from code import interact
import cadnano2.util as util

sys.path.insert(0, os.path.dirname(os.path.abspath(__file__))+"/include")

global sharedApp
sharedApp = None


def ignoreEnv():
    return environ.get('CADNANO_IGNORE_ENV_VARS_EXCEPT_FOR_ME', False)

# The global application object used when cadnano is run as a python module

class HeadlessCadnano(object):
    undoGroup = None
    def isInMaya(self):
        return False
    class prefs():
        squareRows = 50
        squareCols = 50
    def isGui(self):
        return False
# end def

def app(appArgs=None):
    global sharedApp
    if sharedApp != None:
        return sharedApp
    return initAppWithoutGui(appArgs)

def initAppWithoutGui(appArgs=sys.argv):
    global sharedApp
    sharedApp = HeadlessCadnano()
    # loadAllPlugins()
    return sharedApp

def initAppWithGui(appArgs=sys.argv):
    util.qtFrameworkList = ['PyQt', 'PySide']
    from .cadnanoqt import CadnanoQt
    global sharedApp
    sharedApp = CadnanoQt(appArgs)
    sharedApp.finishInit()
    if util.isWindows():
        pass
        # import ctypes
        # myappid = 'harvard.cadnano.cadnano2.2' # arbitrary string
        # ctypes.windll.shell32.SetCurrentProcessExplicitAppUserModelID(myappid)
    return sharedApp
    
def initAppMaya(appArgs=sys.argv):
    import cadnano2.util as util
    util.qtFrameworkList = ['PyQt', 'PySide']
    from cadnanoqt import CadnanoQt
    global sharedApp
    sharedApp = CadnanoQt(appArgs)
    return sharedApp
# end def

def path():
    return os.path.abspath(os.path.dirname(__file__))

# maps plugin path (extension stripped) -> plugin module
loadedPlugins = {}

def unloadedPlugins():
    """ Returns a list of plugin paths that have yet to
    be loaded but are in the top level of one of the
    search directories specified in pluginDirs"""
    internalPlugins = os.path.join(path(), 'plugins')
    pluginDirs = [internalPlugins]
    results = []
    for pluginDir in pluginDirs:
        if not os.path.isdir(pluginDir):
            continue
        for dirent in os.listdir(pluginDir):
            f = os.path.join(pluginDir, dirent)
            isfile = os.path.isfile(f)
            hasValidSuffix = dirent.endswith(('.py', '.so'))
            if isfile and hasValidSuffix:
                results.append(f)
            if os.path.isdir(f) and\
               os.path.isfile(os.path.join(f, '__init__.py')):
                results.append(f)
    return [x for x in results if x not in loadedPlugins]

def loadPlugin(f):
    path, fname = os.path.split(f)
    name, ext = os.path.splitext(fname)
    pluginKey = os.path.join(path, name)
    try:
        mod = loadedPlugins[pluginKey]
        return mod
    except KeyError:
        pass
    file, filename, data = imp.find_module(name, [path])
    mod = imp.load_module(name, file, filename, data)
    loadedPlugins[pluginKey] = mod
    return mod

def loadAllPlugins():
    loadedAPlugin = False
    for p in unloadedPlugins():
        loadPlugin(p)
        loadedAPlugin = True
    return loadedAPlugin


#!/usr/bin/env python
# encoding: utf-8


class StapleExtensionItem(AbstractDecoratorItem):
    def __init__(self, parent):
        """The parent should be a VirtualHelixItem."""
        super(StapleExtensionItem, self).__init__(parent)

    ### SIGNALS ###

    ### SLOTS ###

    ### METHODS ###

    ### COMMANDS ###




class LatticeType:
    Honeycomb = 0
    Square = 1


class EndType:
    FivePrime = 0
    ThreePrime = 1


class StrandType:
    Scaffold = 0
    Staple = 1


class Parity:
    Even = 0
    Odd = 1


class BreakType:
    Left5Prime = 0
    Left3Prime = 1
    Right5Prime = 2
    Right3Prime = 3


class Crossovers:
    honeycombScafLeft = [[1, 11], [8, 18], [4, 15]]
    honeycombScafRight = [[2, 12], [9, 19], [5, 16]]
    honeycombStapLeft = [[6], [13], [20]]
    honeycombStapRight = [[7], [14], [0]]
    squareScafLeft = [[4, 26, 15], [18, 28, 7], [10, 20, 31], [2, 12, 23]]
    squareScafRight = [[5, 27, 16], [19, 29, 8], [11, 21, 0], [3, 13, 24]]
    squareStapLeft = [[31], [23], [15], [7]]
    squareStapRight = [[0], [24], [16], [8]]


class HandleOrient:
    LeftUp = 0
    RightUp = 1
    LeftDown = 2
    RightDown = 3



# Form implementation generated from reading ui file 'dialogs/addseq.ui'
#
# Created by: PyQt6 UI code generator 6.1.1
#
# WARNING: Any manual changes made to this file will be lost when pyuic6 is
# run again.  Do not edit this file unless you know what you are doing.


from PyQt6 import QtCore, QtGui, QtWidgets


class Ui_AddSeqDialog(object):
    def setupUi(self, AddSeqDialog):
        AddSeqDialog.setObjectName("AddSeqDialog")
        AddSeqDialog.resize(500, 500)
        AddSeqDialog.setModal(True)
        self.dialogGridLayout = QtWidgets.QGridLayout(AddSeqDialog)
        self.dialogGridLayout.setObjectName("dialogGridLayout")
        self.tabWidget = QtWidgets.QTabWidget(AddSeqDialog)
        self.tabWidget.setObjectName("tabWidget")
        self.tabStandard = QtWidgets.QWidget()
        self.tabStandard.setObjectName("tabStandard")
        self.standardTabGridLayout = QtWidgets.QGridLayout(self.tabStandard)
        self.standardTabGridLayout.setObjectName("standardTabGridLayout")
        self.groupBox = QtWidgets.QGroupBox(self.tabStandard)
        self.groupBox.setTitle("")
        self.groupBox.setFlat(True)
        self.groupBox.setObjectName("groupBox")
        self.verticalLayout = QtWidgets.QVBoxLayout(self.groupBox)
        self.verticalLayout.setObjectName("verticalLayout")
        self.standardTabGridLayout.addWidget(self.groupBox, 0, 1, 1, 1)
        spacerItem = QtWidgets.QSpacerItem(40, 20, QtWidgets.QSizePolicy.Policy.Expanding, QtWidgets.QSizePolicy.Policy.Minimum)
        self.standardTabGridLayout.addItem(spacerItem, 0, 2, 1, 1)
        spacerItem1 = QtWidgets.QSpacerItem(40, 20, QtWidgets.QSizePolicy.Policy.Expanding, QtWidgets.QSizePolicy.Policy.Minimum)
        self.standardTabGridLayout.addItem(spacerItem1, 0, 0, 1, 1)
        self.tabWidget.addTab(self.tabStandard, "")
        self.tabCustom = QtWidgets.QWidget()
        self.tabCustom.setObjectName("tabCustom")
        self.verticalLayout_2 = QtWidgets.QVBoxLayout(self.tabCustom)
        self.verticalLayout_2.setObjectName("verticalLayout_2")
        self.seqTextEdit = QtWidgets.QTextEdit(self.tabCustom)
        self.seqTextEdit.setObjectName("seqTextEdit")
        self.verticalLayout_2.addWidget(self.seqTextEdit)
        self.tabWidget.addTab(self.tabCustom, "")
        self.dialogGridLayout.addWidget(self.tabWidget, 0, 0, 1, 1)
        self.customButtonBox = QtWidgets.QDialogButtonBox(AddSeqDialog)
        self.customButtonBox.setStandardButtons(QtWidgets.QDialogButtonBox.StandardButton.Apply|QtWidgets.QDialogButtonBox.StandardButton.Cancel)
        self.customButtonBox.setCenterButtons(True)
        self.customButtonBox.setObjectName("customButtonBox")
        self.dialogGridLayout.addWidget(self.customButtonBox, 1, 0, 1, 1)

        self.retranslateUi(AddSeqDialog)
        self.tabWidget.setCurrentIndex(0)
        self.customButtonBox.rejected.connect(AddSeqDialog.reject)
        self.customButtonBox.clicked['QAbstractButton*'].connect(AddSeqDialog.accept)
        QtCore.QMetaObject.connectSlotsByName(AddSeqDialog)
        AddSeqDialog.setTabOrder(self.customButtonBox, self.tabWidget)
        AddSeqDialog.setTabOrder(self.tabWidget, self.seqTextEdit)

    def retranslateUi(self, AddSeqDialog):
        _translate = QtCore.QCoreApplication.translate
        AddSeqDialog.setWindowTitle(_translate("AddSeqDialog", "Choose a sequence"))
        self.tabWidget.setTabText(self.tabWidget.indexOf(self.tabStandard), _translate("AddSeqDialog", "Standard"))
        self.tabWidget.setTabText(self.tabWidget.indexOf(self.tabCustom), _translate("AddSeqDialog", "Custom"))



#!/usr/bin/env python
# encoding: utf-8

import cadnano2.util as util
from cadnano2.views import styles

# import Qt stuff into the module namespace with PySide, PyQt4 independence
util.qtWrapImport('QtCore', globals(), ['QPointF', 'QRectF', 'Qt'])
util.qtWrapImport('QtGui', globals(), ['QBrush',
                                       'QColor',
                                       'QFont',
                                       'QFontMetricsF',
                                       'QPainterPath',
                                       'QPen',
                                       'QTextCursor',
                                       'QTransform'])
util.qtWrapImport('QtWidgets', globals(), ['QGraphicsItem',
                                           'QGraphicsPathItem',
                                           'QGraphicsRectItem',
                                           'QGraphicsTextItem',
                                           'QLabel',
                                           ])


_baseWidth = _bw = styles.PATH_BASE_WIDTH
_halfbaseWidth = _hbw = _baseWidth / 2
_offset1 = _baseWidth / 4
_defaultRect = QRectF(0, 0, _bw, _bw)
_bpen = QPen(styles.bluestroke, styles.INSERTWIDTH)
_rpen = QPen(styles.redstroke, styles.SKIPWIDTH)
_noPen = QPen(Qt.PenStyle.NoPen)


def _insertGen(path, start, c1, p1, c2):
    path.moveTo(start)
    path.quadTo(c1, p1)
    path.quadTo(c2, start)
# end def


_pathStart = QPointF(_hbw, _hbw)
_pathMidUp = QPointF(_hbw, -_bw)
_pathUpUpCtrlPt = QPointF(-_hbw, -_bw)
_pathUpDownCtrlPt = QPointF(1.5 * _bw, -_bw)
_pathMidDown = QPointF(_hbw, 2 * _bw)
_pathDownDownCtrlPt = QPointF(-_hbw, 2 * _bw)
_pathDownUpCtrlPt = QPointF(1.5 * _bw, 2 * _bw)
_insertPathUp = QPainterPath()
_insertGen(_insertPathUp, _pathStart, _pathUpUpCtrlPt,
          _pathMidUp, _pathUpDownCtrlPt)
_insertPathUp.translate(_offset1, 0)
_insertPathUpRect = _insertPathUp.boundingRect()
_insertPathDown = QPainterPath()
_insertGen(_insertPathDown, _pathStart, _pathDownDownCtrlPt,
           _pathMidDown, _pathDownUpCtrlPt)
_insertPathDown.translate(_offset1, 0)
_insertPathDownRect = _insertPathDown.boundingRect()


_bigRect = _defaultRect.united(_insertPathUpRect)
_bigRect = _bigRect.united(_insertPathDownRect)
_bpen2 = QPen(styles.bluestroke, 2)
_offset2 = _bw*0.75
_font = QFont(styles.thefont, 10, QFont.Weight.Bold)
_bigRect.adjust(-15, -15, 30, 30)
# Bases are drawn along and above the insert path.
# These calculations revolve around fixing the characters at a certain
# percentage of the total arclength.
# The fraction of the insert that comes before the first character and
# after the last character is the padding, and the rest is divided evenly.
_fractionInsertToPad = .10


class InsertionPath(object):
    """
    This is just the shape of the Insert item
    """

    def __init__(self):
        super(InsertionPath, self).__init__()
    # end def

    def getPen(self):
        return _bpen
    # end def

    def getInsert(self, istop):
        if istop:
            return _insertPathUp
        else:
            return _insertPathDown
    # end def
# end class


def _xGen(path, p1, p2, p3, p4):
    path.moveTo(p1)
    path.lineTo(p2)
    path.moveTo(p3)
    path.lineTo(p4)
# end def

_pathStart = QPointF(_hbw, _hbw)


class SkipPath(object):
    """
    This is just the shape of the Insert item
    """

    _skipPath = QPainterPath()
    _xGen(_skipPath,
          _defaultRect.bottomLeft(),
          _defaultRect.topRight(),
          _defaultRect.topLeft(),
          _defaultRect.bottomRight())

    def __init__(self):
        super(SkipPath, self).__init__()
    # end def

    def getPen(self):
        return _rpen
    # end def

    def getSkip(self):
        return self._skipPath
    # end def
# end class


_insertPath = InsertionPath()
_skipPath = SkipPath()


class InsertionItem(QGraphicsPathItem):
    """
    This is just the shape of the Insert item
    """
    def __init__(self, virtualHelixItem, strand, insertion):
        super(InsertionItem, self).__init__(virtualHelixItem)
        self.hide()
        self._strand = strand
        self._insertion = insertion
        self._seqItem = QGraphicsPathItem(parent=self)
        self._isOnTop = isOnTop = virtualHelixItem.isStrandOnTop(strand)
        y = 0 if isOnTop else _bw
        self.setPos(_bw*insertion.idx(), y)
        self.setZValue(styles.ZINSERTHANDLE)
        self._initLabel()
        self._initClickArea()
        self.updateItem()
        self.show()
    # end def

    def _initLabel(self):
        """Display the length of the insertion."""
        self._label = label = QGraphicsTextItem("", parent=self)
        label.setFont(_font)
        label.setDefaultTextColor(QColor(self._strand.oligo().color()))
        label.setTextInteractionFlags(Qt.TextInteractionFlag.TextEditorInteraction)
        label.inputMethodEvent = self.inputMethodEventHandler
        label.keyPressEvent = self.textkeyPressEvent
        label.mousePressEvent = self.labelMousePressEvent
        label.mouseDoubleClickEvent = self.mouseDoubleClickEvent
        label.setTextWidth(-1)

        self._label = label
        self._seqItem = QGraphicsPathItem(parent=self)
        self._seqText = None
        self.updateItem()
        self.show()
    # end def

    def _initClickArea(self):
        """docstring for _initClickArea"""
        self._clickArea = cA = QGraphicsRectItem(_defaultRect, self)
        cA.setPen(_noPen)
        cA.mousePressEvent = self.mousePressEvent
        cA.mouseDoubleClickEvent = self.mouseDoubleClickEvent
    # end def

    ### PUBLIC SUPPORT METHODS ###
    def remove(self):
        """
        Called from the following stranditem methods:
            strandRemovedSlot
            strandInsertionRemovedSlot
            refreshInsertionItems
        """
        scene = self.scene()
        self._label.setTextInteractionFlags(Qt.TextInteractionFlag.NoTextInteraction)
        self._label.clearFocus()
        scene.removeItem(self._label)
        self._label = None
        scene.removeItem(self._seqItem)
        self._seqItem = None
        scene.removeItem(self)
        self._insertion = None
        self._strand = None
    # end def

    def updateItem(self):
        self._updatePath()
        self._updateLabel()
        self._updateSequenceText()
        self._resetPosition()
    # end def

    ### PRIVATE SUPPORT METHODS ###
    def _focusOut(self):
        lbl = self._label
        if lbl == None:
            return
        cursor = lbl.textCursor()
        cursor.clearSelection()
        lbl.setTextCursor(cursor)
        lbl.clearFocus()
    # end def

    def _resetPosition(self):
        """
        Set the label position based on orientation and text alignment.
        """
        lbl = self._label
        if lbl == None:
            return
        txtOffset = lbl.boundingRect().width()/2
        insertion = self._insertion
        y = -_bw if self._isOnTop else _bw
        lbl.setPos(_offset2-txtOffset, y)
        if insertion.length() > 0:
            lbl.show()
        else:
            lbl.hide()
    # end def

    def _updateLabel(self):
        self._label.setPlainText("%d" % (self._insertion.length()))
    # end def

    def _updatePath(self):
        strand = self._strand
        if strand == None:
            self.hide()
            return
        else:
            self.show()
        isOnTop = self._isOnTop
        if self._insertion.length() > 0:
            self.setPen(QPen(QColor(strand.oligo().color()), styles.INSERTWIDTH))
            self.setBrush(QBrush(Qt.BrushStyle.NoBrush))
            self.setPath(_insertPath.getInsert(isOnTop))
        else:  # insertionSize < 0 (a skip)
            self.setPen(_skipPath.getPen())
            self.setPath(_skipPath.getSkip())
    # end def

    def setSequence(self, sequence):
        self._seqText = sequence
        self._updateSequenceText()
        self._seqItem.show()
    # end def

    def hideSequence(self):
        self._seqItem.hide()
    # end def

    def _updateSequenceText(self):
        seqItem = self._seqItem
        isOnTop = self._isOnTop
        index = self._insertion.idx()
        baseText = self._seqText
        font = styles.SEQUENCEFONT
        seqFontH = styles.SEQUENCEFONTH
        insertW = styles.INSERTWIDTH
        seqFontCharW = styles.SEQUENCEFONTCHARWIDTH
        # draw sequence on the insert
        if baseText:  # only draw sequences if they exist i.e. not None!
            lenBT = len(baseText)
            if isOnTop:
                angleOffset = 0
            else:
                angleOffset = 180
            if lenBT > 20:
                baseText = baseText[:17] + '...'
                lenBT = len(baseText)
            fractionArclenPerChar = (1.0-2.0*_fractionInsertToPad)/(lenBT+1)
            seqItem.setPen(QPen(Qt.PenStyle.NoPen))
            seqItem.setBrush(QBrush(Qt.GlobalColor.black))

            seqPath = QPainterPath()
            loopPath = self.path()
            for i in range(lenBT):
                frac = _fractionInsertToPad + (i+1)*fractionArclenPerChar
                pt = loopPath.pointAtPercent(frac)
                tangAng = loopPath.angleAtPercent(frac)

                tempPath = QPainterPath()
                # 1. draw the text
                tempPath.addText(0,0, font, baseText[i if isOnTop else -i-1])
                # 2. center it at the zero point different for top and bottom
                # strands
                if not isOnTop:
                    tempPath.translate(0, -seqFontH - insertW)

                tempPath.translate(QPointF(-seqFontCharW/2.,
                                           -2 if isOnTop else seqFontH))
                mat = QTransform()
                # 3. rotate it
                mat.rotate(-tangAng + angleOffset)
                rotatedPath = mat.map(tempPath)
                # 4. translate the rotate object to it's position on the part
                rotatedPath.translate(pt)
                seqPath.addPath(rotatedPath)
            # end for
            seqItem.setPath(seqPath)
        # end if
    # end def

    ### EVENT HANDLERS ###
    def mouseDoubleClickEvent(self, event):
        """Double clicks remove the insertion/skip."""
        self._strand.changeInsertion(self._insertion.idx(), 0)

    def mousePressEvent(self, event):
        """This needs to be present for mouseDoubleClickEvent to work."""
        pass

    def labelMousePressEvent(self, event):
        """
        Pre-selects the text for editing when you click
        the label.
        """
        lbl = self._label
        lbl.setTextInteractionFlags(Qt.TextInteractionFlag.TextEditorInteraction)
        cursor = lbl.textCursor()
        cursor.setPosition(0)
        cursor.movePosition(QTextCursor.MoveOperation.End, mode=QTextCursor.MoveMode.KeepAnchor)
        lbl.setTextCursor(cursor)

    def textkeyPressEvent(self, event):
        """
        Must intercept invalid input events.  Make changes here
        """
        a = event.key()
        text = event.text()
        if a in [Qt.Key.Key_Space, Qt.Key.Key_Tab]:
            return
        elif a in [Qt.Key.Key_Return, Qt.Key.Key_Enter]:
            self.inputMethodEventHandler(event)
            return
        elif str(text).isalpha():
            return
        else:
            return QGraphicsTextItem.keyPressEvent(self._label, event)

    def inputMethodEventHandler(self, event):
        """
        This is run on the label being changed
        or losing focus
        """
        lbl = self._label
        if lbl == None:
            return
        test = str(lbl.toPlainText())
        try:
            insertionSize = int(test)
        except:
            insertionSize = None
        insertion = self._insertion
        length = insertion.length()
        if insertionSize != None and insertionSize != length:
            self._strand.changeInsertion(insertion.idx(), insertionSize)
            if insertion.length():
                self._resetPosition()
        # end if
        self._focusOut()
    # end def
# end class



"""
pathhelixhandle.py
Created by Shawn on 2011-02-05.
"""
from cadnano2.views import styles

import cadnano2.util as util
# import Qt stuff into the module namespace with PySide, PyQt4 independence
util.qtWrapImport('QtCore', globals(), ['QPointF', 'QRectF', 'Qt'])
util.qtWrapImport('QtGui', globals(), ['QBrush',
                                       'QFont',
                                       'QPen',
                                       'QDrag',
                                       'QTransform',
                                       'QUndoCommand'])
util.qtWrapImport('QtWidgets', globals(), ['QGraphicsEllipseItem',
                                           'QGraphicsItem',
                                           'QGraphicsSimpleTextItem',
                                           'QGraphicsTextItem',
                                           'QStyle'])


_radius = styles.VIRTUALHELIXHANDLEITEM_RADIUS
_rect = QRectF(0, 0, 2*_radius + styles.VIRTUALHELIXHANDLEITEM_STROKE_WIDTH,\
        2*_radius + styles.VIRTUALHELIXHANDLEITEM_STROKE_WIDTH)
_defBrush = QBrush(styles.grayfill)
_defPen = QPen(styles.graystroke, styles.VIRTUALHELIXHANDLEITEM_STROKE_WIDTH)
_hovBrush = QBrush(styles.bluefill)
_hovPen = QPen(styles.bluestroke, styles.VIRTUALHELIXHANDLEITEM_STROKE_WIDTH)
_useBrush = QBrush(styles.orangefill)
_usePen = QPen(styles.orangestroke, styles.VIRTUALHELIXHANDLEITEM_STROKE_WIDTH)
_font = styles.VIRTUALHELIXHANDLEITEM_FONT


class VirtualHelixHandleItem(QGraphicsEllipseItem):
    """docstring for VirtualHelixHandleItem"""
    _filterName = "virtualHelix"
    
    def __init__(self, virtualHelix, partItem, viewroot):
        super(VirtualHelixHandleItem, self).__init__(partItem)
        self._virtualHelix = virtualHelix
        self._partItem = partItem
        self._viewroot = viewroot
        self._beingHoveredOver = False
        self.setAcceptHoverEvents(True)
        # handle the label specific stuff
        self._label = self.createLabel()
        self.setNumber()
        self.setFlag(QGraphicsItem.GraphicsItemFlag.ItemIsSelectable)
        self.setFlag(QGraphicsItem.GraphicsItemFlag.ItemSendsScenePositionChanges)
        self.setSelectedColor(False)
        self.setZValue(styles.ZPATHHELIX)
        self.setRect(_rect)
    # end def

    def setSelectedColor(self, value):
        if self.number() >= 0:
            if value == True:
                self.setBrush(_hovBrush)
                self.setPen(_hovPen)
            else:
                self.setBrush(_useBrush)
                self.setPen(_usePen)
        else:
            self.setBrush(_defBrush)
            self.setPen(_defPen)
        self.update(self.boundingRect())
    # end def
    
    def virtualHelix(self):
        return self._virtualHelix

    def paint(self, painter, option, widget):
        painter.setPen(self.pen())
        painter.setBrush(self.brush())
        painter.drawEllipse(self.rect())
    # end def

    def remove(self):
        scene = self.scene()
        scene.removeItem(self._label)
        scene.removeItem(self)
        self._label = None
    # end def

    def someVHChangedItsNumber(self, r, c):
        # If it was our VH, we need to update the number we
        # are displaying!
        if (r,c) == self.vhelix.coord():
            self.setNumber()
    # end def

    def createLabel(self):
        label = QGraphicsSimpleTextItem("%d" % self._virtualHelix.number())
        label.setFont(_font)
        label.setZValue(styles.ZPATHHELIX)
        label.setParentItem(self)
        return label
    # end def

    def setNumber(self):
        """docstring for setNumber"""
        vh = self._virtualHelix
        num = vh.number()
        label = self._label
        radius = _radius

        if num != None:
            label.setText("%d" % num)
        else:
            return
        y_val = radius / 3
        if num < 10:
            label.setPos(radius / 1.5, y_val)
        elif num < 100:
            label.setPos(radius / 3, y_val)
        else: # _number >= 100
            label.setPos(0, y_val)
        bRect = label.boundingRect()
        posx = bRect.width()/2
        posy = bRect.height()/2
        label.setPos(radius-posx, radius-posy)
    # end def

    def number(self):
        """docstring for number"""
        return self._virtualHelix.number()
        
    def partItem(self):
        return self._partItem
    # end def

    def hoverEnterEvent(self, event):
        """
        hoverEnterEvent changes the PathHelixHandle brush and pen from default
        to the hover colors if necessary.
        """
        if not self.isSelected():
            if self.number() >= 0:
                if self.isSelected():
                    self.setBrush(_hovBrush)
                else:
                    self.setBrush(_useBrush)
            else:
                self.setBrush(_defBrush)
            self.setPen(_hovPen)
            self.update(self.boundingRect())
    # end def

    def hoverLeaveEvent(self, event):
        """
        hoverEnterEvent changes the PathHelixHanle brush and pen from hover
        to the default colors if necessary.
        """
        if not self.isSelected():
            self.setSelectedColor(False)
            self.update(self.boundingRect())
    # end def

    def mousePressEvent(self, event):
        """
        All mousePressEvents are passed to the group if it's in a group
        """
        selectionGroup = self.group()
        if selectionGroup != None:
            selectionGroup.mousePressEvent(event)
        else:
            QGraphicsItem.mousePressEvent(self, event)
    # end def

    def mouseMoveEvent(self, event):
        """
        All mouseMoveEvents are passed to the group if it's in a group
        """
        selectionGroup = self.group()
        if selectionGroup != None:
            selectionGroup.mousePressEvent(event)
        else:
            QGraphicsItem.mouseMoveEvent(self, event)
    # end def

    def restoreParent(self, pos=None):
        """
        Required to restore parenting and positioning in the partItem
        """

        # map the position
        partItem = self._partItem
        if pos == None:
            pos = self.scenePos()
        self.setParentItem(partItem)
        self.setSelectedColor(False)

        assert(self.parentItem() == partItem)
        # print "restore", self.number(), self.parentItem(), self.group()
        assert(self.group() == None)
        tempP = partItem.mapFromScene(pos)
        self.setPos(tempP)
        self.setSelected(False)
    # end def

    def itemChange(self, change, value):
        # for selection changes test against QGraphicsItem.GraphicsItemChange.ItemSelectedChange
        # intercept the change instead of the has changed to enable features.

        if change == QGraphicsItem.GraphicsItemChange.ItemSelectedChange and self.scene():
            viewroot = self._viewroot
            currentFilterDict = viewroot.selectionFilterDict()
            selectionGroup = viewroot.vhiHandleSelectionGroup()

            # only add if the selectionGroup is not locked out
            if value == True and self._filterName in currentFilterDict:
                if self.group() != selectionGroup:
                    selectionGroup.pendToAdd(self)
                    selectionGroup.setSelectionLock(selectionGroup)
                    self.setSelectedColor(True)
                    return True
                else:
                    return False
            # end if
            elif value == True:
                # don't select
                return False
            else:
                # Deselect
                selectionGroup.pendToRemove(self)
                self.setSelectedColor(False)
                return False
            # end else
        # end if
        return QGraphicsEllipseItem.itemChange(self, change, value)
    # end def
    
    def modelDeselect(self, document):
        pass
        self.restoreParent()
    # end def
    
    def modelSelect(self, document):
        pass
        self.setSelected(True)
    # end def
# end class
